# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports
from django.db import models

# iop4lib imports
from ..enums import *
from iop4lib.telescopes import Telescope
from iop4lib.instruments import Instrument
from .fitfilemodel import FitFileModel
from .fields import FlagChoices, FlagBitField

# other imports
import re
import os
import stat
import datetime
import numpy as np

# logging 
import logging
logger = logging.getLogger(__name__)



class RawFit(FitFileModel):

    # Database fields and information

    ## Identifiers
    
    epoch = models.ForeignKey('Epoch', on_delete=models.CASCADE, related_name="%(class)ss") # rawfits
    filename = models.CharField(max_length=255)

    # flags

    class FLAGS(FlagChoices):
        ERROR = 1 << 1
        ERROR_DOWNLOAD = 1 << 2
        DOWNLOADED = 1 << 3
        ERROR_CLASSIFY = 1 << 4
        CLASSIFIED = 1 << 5
        ERROR_ASTROMETRY = 1 << 6
        BUILT_REDUCED = 1 << 7
        
    flags = FlagBitField(choices=FLAGS.choices(), default=0)

    ## FITS related fields

    instrument = models.CharField(max_length=20, choices=INSTRUMENTS.choices, default=INSTRUMENTS.NONE, help_text="Instrument used for the observation.")
    obsmode = models.CharField(max_length=20, choices=OBSMODES.choices, default=OBSMODES.NONE, help_text="Whether the observation was photometry or polarimetry.")
    juliandate = models.FloatField(null=True, help_text="Julian date of observation, from the date and time indicated in the FITS header.")
    imgsize = models.CharField(max_length=12, null=True, help_text="String with the format _width_x_height_ for the image, as in 1024x1024.") # as in 1024x1024, 2048x2048, etc
    imgtype = models.CharField(max_length=20, choices=IMGTYPES.choices, default=IMGTYPES.NONE, help_text="Whether the raw image is a bias (BIAS), flat (FLAT), or science (LIGHT).")
    band = models.CharField(max_length=10, choices=BANDS.choices, default=BANDS.NONE, help_text="Band of the observation, as in the filter used (R, V, etc).")
    rotangle = models.FloatField(null=True, help_text="Rotation angle of the polarizer in degrees.")
    exptime = models.FloatField(null=True, help_text="Exposure time in seconds.")

    # DB information and constraints

    class Meta:
        app_label = 'iop4api'
        verbose_name = "RawFit"
        verbose_name_plural = "Raw FITS files"
        constraints = [
            models.UniqueConstraint(fields=['epoch', 'filename'], name='epoch_filename_unique')
        ]

    # Properties 

    @property
    def filepath(self):
        """ Returns the full path to the FITS file.
        The raw FITS file are stored in the rawfitsdir of the epoch.
        """
        return os.path.join(self.epoch.rawfitsdir, self.filename)
    
    @property
    def filedpropdir(self):
       """ Returns the full path to the directory where related files are stored.
       The directory is in the calibrationdir of the epoch, different from the rawfitsfit,
       to keep a clean archive of raw data (see the `Epoch` class). Reduced FITS files will be
       stored here also, see the `ReducedFit` class.
       """
       return os.path.join(self.epoch.calibrationdir, 
                            self.filename +  '.d', 
                            self.__class__.__name__)
    
    @property
    def fileloc(self):
        """ Returns an human readable string which uniquely identifies the file, as an alternative to the id."""
        return f"{self.epoch.epochname}/{self.filename}"
    
    @property
    def telescope(self):
        return self.epoch.telescope
    
    @property
    def night(self):
        return self.epoch.night


    # repr and str

    def __repr__(self):
        return f"{self.__class__.__name__}.objects.get(id={self.id!r})"
    
    def __str__(self):
        return f"<{self.__class__.__name__} {self.id!r} | {self.fileloc}>"

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text(f"{self!r}")
        else:
            with p.group(4, f'<{self.__class__.__name__}(', ')>'):
                p.text(f"id: {self.id},")
                p.breakable()
                p.text(f"telescope: {self.epoch.telescope},")
                p.breakable()
                p.text(f"night: {self.epoch.night},")
                p.breakable()
                p.text(f"filename: {self.filename},")
                p.breakable()
                p.text(f"instrument: {self.instrument},")
                p.breakable()
                p.text(f"imgtype: {self.imgtype},")
                p.breakable()
                p.text(f"size: {self.imgsize},")
                p.breakable()
                p.text(f"obsmode: {self.obsmode},")
                p.breakable()
                p.text(f"rotangle: {self.rotangle},")
                p.breakable()
                p.text(f"band: {self.band},")
                p.breakable()
                p.text(f"exptime: {self.exptime},")
                p.breakable()
                p.text(f"flags: {', '.join(self.flag_labels)}")

    def _repr_html_(self):
        return (f"{self.__class__.__name__}(id={self.id!r}):<br>\n"
                f" - telescope: {self.epoch.telescope}<br>\n"
                f" - night: {self.epoch.night}<br>\n"
                f" - filename: {self.filename}<br>\n"
                f" - instrument: {self.instrument}<br>\n"
                f" - imgtype: {self.imgtype}<br>\n"
                f" - size: {self.imgsize}<br>\n"
                f" - obsmode: {self.obsmode}<br>\n"
                f" - band: {self.band}<br>\n"
                f" - exptime: {self.exptime}<br>\n"
                f" - flags: {(','.join(self.flag_labels))}<br>\n")     
    
    # Constructor

    @staticmethod 
    def fileloc_to_tel_night_filename(fileloc):
        """Parses a fileloc to telescope, night and filename."""
        from .epoch import Epoch

        matches = re.findall(r"(([a-zA-Z0-9]+)/([0-9]{2,4}-?[0-9]{2}-?[0-9]{2}))/([^/\\]+)$", fileloc)

        if len(matches) != 1:
            raise Exception(f"fileloc {fileloc} is not EPOCHNAME/filename")
        
        epochname = matches[0][0]

        telescope, night = Epoch.epochname_to_tel_night(epochname)
        
        filename = matches[0][-1]

        return telescope, night, filename
      
    @classmethod
    def create(cls, 
                 epoch=None, filename=None, fileloc=None,
                 auto_procure_local=True,
                 reset_flags=False, 
                 auto_classify=True,
                 force_redownload=False,
                 auto_merge_to_db=True):
        """
            Create a RawFit object from fileloc or epoch and filename.

            Flags are cleared when .create() is called.

            Parameters
            ----------
            fileloc : str
                File location in the format TELESCOPE/YYYYMMDD/FILENAME.
            epoch : Epoch
                Epoch object.
            filename : str
                Filename.
            auto_procure_local : bool
                Whether automatically procure the local file (linking it to a local file).
            auto_classify : bool
                If auto_procure_local is True, whether to automatically classify the file.
            force_redownload : bool
                If auto_procure_local is True, whether to force redownload of the file even if it already exists.
            auto_merge_to_db : bool
                Whether automatically merge the object to the DB when making any changes, including creation.
        """

        from iop4lib.db import Epoch

        if epoch is not None and filename is not None:
            fileloc = os.path.join(epoch.epochname, filename)
        elif fileloc is not None: # get epoch and filename from fileloc

            telescope, night, filename = RawFit.fileloc_to_tel_night_filename(fileloc)

            if (epoch := Epoch.objects.filter(telescope=telescope, night=night).first()) is  None:
                logger.warning("No epoch in DB; it will be created.")
                epoch = Epoch.create(telescope=telescope, night=night, auto_merge_to_db=auto_merge_to_db, auto_link_rawfits=auto_procure_local)
            
            logger.info(f"Setting epoch to {epoch}.")
        else:
            raise Exception("Either epoch and filename or fileloc must be provided.")

        # create entry

        if (rawfit := RawFit.objects.filter(epoch=epoch, filename=filename).first()) is None:
            logger.debug(f"Creating DB entry for {fileloc}.")
            rawfit = cls(epoch=epoch, filename=filename)
            rawfit.save()
        else:
            logger.debug(f"DB entry for {fileloc} already exists, it will be used instead.")

        # intance only attributes
        rawfit.auto_merge_to_db = auto_merge_to_db
        rawfit.auto_classify = auto_classify
        rawfit.auto_procure_local = auto_procure_local
        rawfit.reset_flags = reset_flags
        rawfit.force_redownload = force_redownload

        # link to local file
        if auto_procure_local:
            rawfit.procure_local_file()

        # merge to db
        if auto_merge_to_db:
            rawfit.save()

        return rawfit
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.auto_procure_local=True
        self.auto_classify=True
        self.reset_flags=False
        self.force_redownload=False
        self.auto_merge_to_db=True
    
    # Methods

    def download(self):
        if not os.path.isdir(self.epoch.rawfitsdir):
            logger.debug("Creating rawfitsdir")
            os.makedirs(self.epoch.rawfitsdir)
        
        Telescope.by_name(self.epoch.telescope).download_rawfits([self])
               
    def procure_local_file(self):

        logger.debug(f"{self}: procuring local file.")

        if self.fileexists and self.reset_flags:
            self.clear_flags()

        if not self.fileexists or self.force_redownload:
            try:
                logger.debug(f"Downloading file {self.fileloc}.")
                self.download()
                self.set_flag(RawFit.FLAGS.DOWNLOADED)
            except Exception as e:
                logger.error(f"Error downloading file {self.filename}, skipping download: {e}.")
                self.set_flag(RawFit.FLAGS.ERROR_DOWNLOAD)
        else:
            self.set_flag(RawFit.FLAGS.DOWNLOADED)

        if self.fileexists:
            if iop4conf.set_rawdata_readonly:
                # this will remove the write permission from the file for all users
                current_perms = os.stat(self.filepath).st_mode
                new_perms = current_perms & (~stat.S_IWUSR) & (~stat.S_IWGRP) & (~stat.S_IWOTH)
                os.chmod(self.filepath, new_perms)

            if self.auto_classify:
                self.classify()

        if self.auto_merge_to_db:
            self.save()
        
    def classify(self):
        """
            Get some parameters from the fit file to the DB entry.
        """
        logger.debug(f"{self}: classifying")

        try:
            self.unset_flag(RawFit.FLAGS.ERROR_CLASSIFY)
            self.unset_flag(RawFit.FLAGS.CLASSIFIED)
            Telescope.by_name(self.epoch.telescope).classify_rawfit(self)
            self.set_flag(RawFit.FLAGS.CLASSIFIED)
        except Exception as e:
            logger.error(f"Error classifying {self.fileloc}: {e}.")
            self.set_flag(RawFit.FLAGS.ERROR_CLASSIFY)

        if self.auto_merge_to_db:
            self.save()

    def request_master(self, model, other_epochs=False):
        return Instrument.by_name(self.instrument).request_master(self, model, other_epochs=other_epochs)

    def request_masterbias(self, *args, **kwargs):
        from iop4lib.db import MasterBias
        return self.request_master(MasterBias, *args, **kwargs)
    
    def request_masterflat(self, *args, **kwargs):
        from iop4lib.db import MasterFlat
        return self.request_master(MasterFlat, *args, **kwargs)
    
    def request_masterdark(self, *args, **kwargs):
        from iop4lib.db import MasterDark
        return self.request_master(MasterDark, *args, **kwargs)

    @property
    def header_hintcoord(self):
        """ Returns a SkyCoord according to the headers of the FITS file."""
        return Instrument.by_name(self.instrument).get_header_hintcoord(self)

    @property
    def header_hintobject(self):
        """ Returns the AstroSource according to the OBJECT keyword in the header of the FITS file. """
        return Instrument.by_name(self.instrument).get_header_hintobject(self)
    
    # Class methods    

    @classmethod
    def by_fileloc(cls, fileloc):
        telescope, night, filename = cls.fileloc_to_tel_night_filename(fileloc)
        return cls.objects.filter(epoch__telescope=telescope, epoch__night=night, filename=filename).get()


#!/usr/bin/env python
""" iop4_daily_summary.py
    Generate and send a summary of IOP4 results for a given night.

Contact: Juan Escudero Pedrosa (jescudero@iaa.es).
"""

# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=True)

# django imports
from django.template import Context, Template

# iop4lib imports
from iop4lib.db import Epoch, RawFit, ReducedFit, PhotoPolResult, AstroSource
from iop4lib.instruments import Instrument
from iop4lib.enums import SRCTYPES
from iop4lib.utils import get_column_values

# other imports
import os, sys
from pathlib import Path
import argparse
import datetime
import io
import base64
import numpy as np
import scipy as sp
import matplotlib as mplt
import matplotlib.pyplot as plt
from astropy.time import Time
import smtplib, email

# logging
import coloredlogs, logging
coloredlogs.install(level='INFO')
logger = logging.getLogger(__name__)



def gather_context(args):
    r""" Build the context for the night summary template."""

    context = Context()
    context['night'] = args.date

    # List epochs for the given date
    # annotate the number of distinct sources in each epoch
    
    epochs = list(Epoch.objects.filter(night=args.date))

    # annotate epochs with some useful information

    for epoch in epochs:
        # annotate with the set of (non-calibrator) sources in the results each epoch

        epoch.srcname_list = list(PhotoPolResult.objects.filter(epoch=epoch)
                                  .exclude(astrosource__in=AstroSource.objects.filter(is_calibrator=True))
                                  .values_list("astrosource__name", flat=True)
                                  .distinct())
        
        # annotate with the list of files with ERROR_ASTROMETRY

        epoch.files_with_error_astrometry = list(ReducedFit.objects.filter(epoch=epoch)
                                                  .filter(flags__has=ReducedFit.FLAGS.ERROR_ASTROMETRY))
        epoch.files_with_error_astrometry_ids_str_L = [str(reducedfit.id) for reducedfit in epoch.files_with_error_astrometry]
        epoch.files_with_error_astrometry_href = args.site_url + "/iop4/admin/iop4api/reducedfit/?id__in=" + ",".join(epoch.files_with_error_astrometry_ids_str_L)
        # TODO: include IOP4Admin in Django configure and use reverse as in PhotoPolResult admin for line above

    # get list of all sources for which there are results in this night
        
    sources = AstroSource.objects.exclude(is_calibrator=True).filter(photopolresults__epoch__night=args.date).distinct()

    if sources:
        instruments = [instrument.name for instrument in Instrument.get_known()]
        colors = [mplt.colormaps['tab10'](i) for i in range(len(instruments))]

    results_summary_images = dict()

    for source in sources:

        logger.info(f"Plotting summary of results for source {source.name}")

        # check which was the previous night with results for this source

        prev_night = (PhotoPolResult.objects
                      .filter(astrosource=source, band="R", epoch__night__lt=args.date)
                      .order_by('-epoch__night')
                      .values_list('epoch__night', flat=True)
                      .last())

        # get the results for this source from the previous night, or for this night if there is no previous night

        if prev_night is not None:
            qs0 = PhotoPolResult.objects.filter(astrosource=source, band="R").filter(epoch__night__gte=args.date).order_by('-juliandate')
        else:
            qs0 = PhotoPolResult.objects.filter(astrosource=source, band="R").filter(epoch__night=args.date).order_by('-juliandate')


        fig = mplt.figure.Figure(figsize=(800/100, 600/100), dpi=100)
        axs = fig.subplots(nrows=3, ncols=1, sharex=True, gridspec_kw={'hspace': 0.05})

        for instrument, color in zip(instruments, colors):
            qs = qs0.filter(instrument=instrument)

            if not qs.exists():
                continue

            column_names = ['id', 'juliandate', 'band', 'instrument', 'mag', 'mag_err', 'p', 'p_err', 'chi', 'chi_err', 'flags']
            vals =  get_column_values(qs, column_names)
            vals['datetime'] = Time(vals['juliandate'], format='jd').datetime

            axs[0].errorbar(x=vals['datetime'], y=vals['mag'], yerr=vals['mag_err'], marker=".", color=color, linestyle="none")
            axs[1].errorbar(x=vals['datetime'], y=vals['p'], yerr=vals['p_err'], marker=".", color=color, linestyle="none")
            axs[2].errorbar(x=vals['datetime'], y=vals['chi'], yerr=vals['chi_err'], marker=".", color=color, linestyle="none")

        # invert magnitude axis
        axs[0].invert_yaxis()

        # secondary axes and its x limits
        datetime_min, datetime_max = mplt.dates.num2date(axs[0].get_xlim())
        mjd_min, mjd_max = Time([datetime_min, datetime_max]).mjd
        ax0_2 = axs[0].twiny()
        ax0_2.set_xlim([mjd_min, mjd_max])
        ax0_2.ticklabel_format(useOffset=False)

        # major x and secondary x ticks
        ax0_2.xaxis.set_major_locator(mplt.ticker.MaxNLocator(5))
        
        # x and secondary x labels
        axs[-1].set_xlabel('date')
        ax0_2.set_xlabel('MJD')

        # y labels
        axs[0].set_ylabel(f"mag (R)")
        axs[1].set_ylabel("p")
        axs[1].yaxis.set_major_formatter(mplt.ticker.PercentFormatter(1.0, decimals=1))
        axs[2].set_ylabel("chi [º]")

        # title 
        fig.suptitle(f"{source.name} ({source.other_names})" if source.other_names else source.name)

        # legend
        legend_handles = [axs[0].plot([],[],color=color, marker=".", linestyle="none", label=instrument)[0] for color, instrument in zip(colors, instruments)]
        fig.legend(handles=legend_handles, ncols=3, loc='upper left', bbox_to_anchor=(0, 0))

        # save the figure to a buffer and encode it to base64
        buf = io.BytesIO()
        fig.savefig(buf, format='png', bbox_inches='tight', pad_inches=0)
        fig.clf()
        buf.seek(0)
        imgbytes = buf.read()
        imgb64 = base64.b64encode(imgbytes).decode("utf-8")

        results_summary_images[source.name] = imgb64

    # save vars to context and return it
    
    context['epochs'] = epochs    
    context['sources'] = sources
    context['results_summary_images'] = results_summary_images
    context['args'] = args

    return context
 


def generate_html_summary(args, context):
    """ Render the summary template with the given context."""
    
    with open(Path(__file__).parent / 'iop4_night_summary.html', 'r') as f:
        template = Template(f.read())
        return template.render(context=context)



def send_email(args, summary_html):
    """ Send the html summary by email."""

    msg = email.message.EmailMessage()

    msg['Subject'] = f"IOP4 summary {args.date}"
    msg['From'] = args.fromaddr 
    msg['To'] = args.mailto
    msg['reply-to'] = args.contact_email
    msg.set_content(summary_html, subtype="html")

    logger.debug("Sending email")

    with smtplib.SMTP('localhost') as s:
        s.send_message(msg)



def main():
    argparser = argparse.ArgumentParser(description='IOP4 night summary',
                                        prog="iop4-night-summary",
                                        epilog = __doc__,
                                        formatter_class=argparse.RawDescriptionHelpFormatter,
                                        allow_abbrev=False)
    
    argparser.add_argument('--date', type=str, default=None, help='Date of the night in format YYYY-MM-DD (default: yesterday)')
    argparser.add_argument('--mailto', type=lambda s: s.split(','), default=None, help='One or several email addresses separated by commas')
    argparser.add_argument('--fromaddr', type=str, default="iop4@localhost", help='Email address of the sender')
    argparser.add_argument('--contact-name', type=str, default=None, help='Name to indicate as contact, if any')
    argparser.add_argument('--contact-email', type=str, default=None, help='Email to indicate as contact (default is the sender address)')
    argparser.add_argument('--site-url', type=str, default="localhost:8000", help='URL of the site to link the summary')
    argparser.add_argument('--saveto', type=str, default=None, help='Save the summary to a file')
    argparser.add_argument('--rc', type=int, default=None, help="Indicate the return code from IOP4 to warn the user if the data processing fails")

    args = argparser.parse_args()

    if args.date is None:
        args.date = datetime.date.today() - datetime.timedelta(days=1)
    else:
        try:
            args.date = datetime.datetime.strptime(args.date, '%Y-%m-%d').date()
        except ValueError:
            print('Invalid date format. Use YYYY-MM-DD.')
            sys.exit(-1)

    if args.contact_email is None:
        args.contact_email = args.fromaddr
    
    logger.info(f'Generating daily summary for {args.date}')
    context = gather_context(args)   

    logger.info('Rendering html summary')
    html_summary = generate_html_summary(args, context)

    if args.mailto:
        logger.info('Sending email')
        send_email(args, html_summary) 
    else:
        logger.warning('No email address provided.')

    if args.saveto:
        logger.info(f'Saving summary to {args.saveto}')
        with open(args.saveto, 'w') as f:
            f.write(html_summary)
    else:
        logger.warning('No file provided to save the summary.')

    logger.info('Done')

    sys.exit(0)

if __name__ == '__main__':
    main()



import itertools
import numpy as np

def distance(h1,h2):
    return np.linalg.norm(np.array(h1)-np.array(h2))
    
def coords_astrometry(A,B,C,D):
    
    A, B, C, D = A-A, B-A, C-A, D-A
    
    X = (B + np.array([[0,1],[-1,0]]) @ B) / 2
    Y = (B + np.array([[0,-1],[+1,0]]) @ B) / 2

    xc, yc = np.dot(C,X) / np.linalg.norm(C) / np.linalg.norm(X), np.dot(C,Y) / np.linalg.norm(C) / np.linalg.norm(Y)
    xd, yd = np.dot(D,X) / np.linalg.norm(D) / np.linalg.norm(X), np.dot(D,Y) / np.linalg.norm(D) / np.linalg.norm(Y)

    return np.array([xc,yc,xd,yd])
        
def hash_astrometry(points):
    max_distance = -1
    farthest_pair = None
    farthest_i_j = None, None
    
    for i, point1 in enumerate(points):
        for j, point2 in enumerate(points):
            if i >= j:
                continue
            dist = distance(point1, point2)
            if dist > max_distance:
                max_distance = dist
                farthest_pair = (point1, point2)
                farthest_i_j = i, j

    i, j = farthest_i_j
    
    A = np.array(points[i])
    B = np.array(points[j])
    
    k, l = [m for m in [0,1,2,3] if m not in [i,j]]
    
    C = np.array(points[k])
    D = np.array(points[l])

    xc,yc,xd,yd = coords_astrometry(A,B,C,D)

    # this is the astrometry.net invariant (under scaling, rotation and translation)

    if not xc <= xd:
        A, B = B, A
        xc,yc,xd,yd = coords_astrometry(A,B,C,D)
        
    if not xc+xd <= 1:
        C, D = D, C
        xc,yc,xd,yd = coords_astrometry(A,B,C,D)

    return xc,yc,xd,yd
    
def order(points):
    # Calculate the centroid of the points
    cx = np.mean([p[0] for p in points])
    cy = np.mean([p[1] for p in points])

    # Sort the points based on the angle with respect to the centroid
    sorted_points = sorted(points, key=lambda p: np.arctan2(p[1]-cy, p[0]-cx))

    return sorted_points

def hash_ish_old(points):
    P1,P2,P3,P4 = points     
    P1,P2,P3,P4 = order(points)
    d1,d2,d3,d4 = map(np.linalg.norm, [P2-P1,P3-P2,P4-P3,P1-P4])
    return d1,d2,d3,d4



def quad_coords_ish(A,B,C,D):

    (A,B,C,D), idx = force_AB_maxdist([A,B,C,D])

    P = (A+B)/2
    A, B, C, D = A-P,B-P,C-P,D-P
    
    X = B / np.linalg.norm(B)
    Y = np.array([[0,-1],[+1,0]]) @ X

    xa, ya = np.dot(A,X), np.dot(A,Y)
    xb, yb = np.dot(B,X), np.dot(B,Y)
    xc, yc = np.dot(C,X), np.dot(C,Y)
    xd, yd = np.dot(D,X), np.dot(D,Y)

    A = np.array([xa,ya])
    B = np.array([xb,yb])
    C = np.array([xc,yc])
    D = np.array([xd,yd])

    FX = np.array([[-1,0],[0,1]])
    FY = np.array([[1,0],[0,-1]])

    # begin track the idx

    if C[0] + D[0] > 0:
        idx[2], idx[3] = idx[3], idx[2]
        idx[0], idx[1] = idx[1], idx[0]
    if C[0] > D[0]:
        idx[2], idx[3] = idx[3], idx[2]
    
    # end track the idx

    if C[0] + D[0] > 0:
        C,D = [FX@P for P in [C,D]]

    if C[1] + D[1] > 0:
        C,D = [FY@P for P in [C,D]]

    if C[0] > D[0]:
        C,D = D,C

    return (A,B,C,D), idx


def force_AB_maxdist(points):
    """ Given points = A,B,C,D,... reorders then and returns A',B',C',... A' and B' are the maximum distance points."""
    
    max_distance = -1
    farthest_i_j = None, None
    
    for i, point1 in enumerate(points):
        for j, point2 in enumerate(points):
            if i >= j:
                continue
            dist = distance(point1, point2)
            if dist > max_distance:
                max_distance = dist
                farthest_i_j = i, j

    i, j = farthest_i_j
    
    result_idx = [i,j]
    result = [points[i], points[j]]

    result_idx.extend([k for k in range(len(points)) if k not in [i,j]])
    result.extend([points[k] for k in range(len(points)) if k not in [i,j]])
    
    return result, result_idx
    

def hash_ish(points):
    
    A,B,C,D = points
    (A,B,C,D), idx = quad_coords_ish(A,B,C,D)
    d1,d2,d3,d4 = map(np.linalg.norm, [C-A,D-C,B-D,A-B])

    return d1,d2,d3,d4
    
    
def qorder_ish(points):

    (Ap,Bp,Cp,Dp), idx_quad_ord = quad_coords_ish(points[0],points[1],points[2],points[3])

    pts_quad_ord = [points[i] for i in idx_quad_ord]

    # cx = np.mean([p[0] for p in [Ap,Bp,Cp,Dp]])
    # cy = np.mean([p[1] for p in [Ap,Bp,Cp,Dp]])
    # idx_sorting, _ = list(zip(*sorted(enumerate([Ap,Bp,Cp,Dp]), key=lambda p: np.arctan2(p[1][1]-cy, p[1][0]-cx))))
    # A,B,C,D = [pts_quad_ord[i] for i in idx_sorting]

    A,B,C,D = pts_quad_ord

    return A,B,C,D


def find_linear_transformation(P1, P2):
    P1, P2 = np.array(P1), np.array(P2)
    
    # Center points
    P1_mean, P2_mean = np.mean(P1, axis=0), np.mean(P2, axis=0)
    P1_centered, P2_centered = P1 - P1_mean, P2 - P2_mean
    
    # SVD
    H = P1_centered.T @ P2_centered
    U, _, Vt = np.linalg.svd(H)
    
    # Rotation + Scaling Matrix
    R = Vt.T @ U.T
    
    # Translation Vector
    t = P2_mean - R @ P1_mean
    
    return R, t

def find_best_transformation(P1, P2, distance_function):
    """Find the linear transformation of P1 to P2 that minimizes the distance function, searching over all permutations of P2."""

    P1, P2 = np.array(P1), np.array(P2)
    
    # Store all permutations and their corresponding transformations (R, t)
    permutations = list(itertools.permutations(P2))
    transformations = [find_linear_transformation(P1, perm) for perm in permutations]
    
    # Compute distances to the target transformation (using distance_function)
    distances = [distance_function(R) for R, _ in transformations]
    
    # Find the permutation with the minimum distance
    best_index = np.argmin(distances)
    
    best_R, best_t = transformations[best_index]
    best_perm = permutations[best_index]
    
    return best_R, best_t, best_perm

def distance_to_y_flip(R):
    # Target matrix for a single Y-axis flip
    R_target = np.array([[1, 0], [0, -1]])
    
    # Compute Frobenius norm of the difference
    return np.linalg.norm(R - R_target)

def distance_to_identity(R):
    # Target matrix is the identity matrix
    R_target = np.eye(2)  # Identity matrix of size 2x2
    
    # Compute Frobenius norm of the difference
    return np.linalg.norm(R - R_target)


"""
Sphinx plugin to run generate a gallery for notebooks

Adapted from [pymc](https://github.com/pymc-devs/pymc-examples/blob/d5e25ca283ab015c8eded9f82c3df5f68de84314/sphinxext/gallery_generator.py):
    Modified from the seaborn project, which modified the mpld3 project.
"""

import base64
import json
import os
import shutil

import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt

from matplotlib import image
from glob import glob
from pathlib import Path

DOC_SRC = Path(__file__).parent.parent.resolve().absolute()
DEFAULT_IMG_LOC = DOC_SRC / "_static" / "jupyter.png"
THUMBNAIL_DIR = DOC_SRC / "_build" / "html" / "_thumbnails"

if not THUMBNAIL_DIR.exists():
    THUMBNAIL_DIR.mkdir(parents=True)

def create_thumbnail(infile, width=275, height=275, cx=0.5, cy=0.5, border=4):
    """Overwrites `infile` with a new file of the given size"""
    im = image.imread(infile)
    rows, cols = im.shape[:2]
    size = min(rows, cols)
    if size == cols:
        xslice = slice(0, size)
        ymin = min(max(0, int(cx * rows - size // 2)), rows - size)
        yslice = slice(ymin, ymin + size)
    else:
        yslice = slice(0, size)
        xmin = min(max(0, int(cx * cols - size // 2)), cols - size)
        xslice = slice(xmin, xmin + size)
    thumb = im[yslice, xslice]
    thumb[:border, :, :3] = thumb[-border:, :, :3] = 0
    thumb[:, :border, :3] = thumb[:, -border:, :3] = 0

    dpi = 100
    fig = plt.figure(figsize=(width / dpi, height / dpi), dpi=dpi)

    ax = fig.add_axes([0, 0, 1, 1], aspect="auto", frameon=False, xticks=[], yticks=[])
    ax.imshow(thumb, aspect="auto", resample=True, interpolation="bilinear")
    fig.savefig(infile, dpi=dpi)
    plt.close(fig)
    return fig


class NotebookGenerator:
    """Tools for generating an example page from a file"""

    def __init__(self, filename):
        self.basename = os.path.basename(filename)
        stripped_name = os.path.splitext(self.basename)[0]
        self.image_dir = THUMBNAIL_DIR
        self.png_path = os.path.join(self.image_dir, f"{stripped_name}.png")
        try:
            with open(filename) as fid:
                self.json_source = json.load(fid)
        except FileNotFoundError:
            print(f"File {filename} not found")
        self.default_image_loc = DEFAULT_IMG_LOC

    def extract_preview_pic(self):
        """By default, just uses the last image in the notebook."""
        pic = None
        for cell in self.json_source["cells"]:
            for output in cell.get("outputs", []):
                if "image/png" in output.get("data", []):
                    pic = output["data"]["image/png"]
        if pic is not None:
            return base64.b64decode(pic)
        return None

    def gen_previews(self):
        preview = self.extract_preview_pic()
        if preview is not None:
            with open(self.png_path, "wb") as buff:
                buff.write(preview)
        else:
            print(f"I can't find a thumbnail for {self.png_path}")
            shutil.copy(self.default_image_loc, self.png_path)
        create_thumbnail(self.png_path)


def main(app, e):

    nb_paths = glob("recipes/*.py")
    nb_names = [Path(nb_path).with_suffix('.ipynb').name for nb_path in nb_paths]
    ipynb_paths = [str(DOC_SRC / "_build" / "jupyter_execute" / "recipes" / nb_name) for nb_name in nb_names]

    for ipynb_path in ipynb_paths:
        nbg = NotebookGenerator(ipynb_path)
        nbg.gen_previews()


def setup(app):
    app.connect("build-finished", main)


# %% [markdown]
# # Quad matching in DIPOL polarimetry images

# %%
# %autoawait off
# %matplotlib inline
# %run 01_notebook_configuration.py

# To reduce disk usage, only a central cut of of the full field of the DIPOL 
# camera is actually saved.

# %%
# polarimetry (cut)
from IPython.display import display
from iop4lib.db import RawFit, ReducedFit
from iop4lib.utils.plotting import imshow_w_sources
rf_pol = RawFit.by_fileloc(f"OSN-T090/2023-11-06/BLLAC_R_IAR-0760.fts")
display(rf_pol)
imshow_w_sources(rf_pol.mdata)

# %% [markdown]
# The position of the cut is defined in the header of the raw file

# %%
rf_pol.header['XORGSUBF'], rf_pol.header['YORGSUBF']

# %% [markdown]
# We can compare it with the full field image of the photometry field

# %%
# photometry filed (full field)
rf_phot = RawFit.by_fileloc(f"OSN-T090/2023-11-06/BLLac_IAR-0001R.fit")
display(rf_phot)
imshow_w_sources(rf_phot.mdata)

# %% [markdown]
# Let's build the photometry field first

# %% 
redf_phot = ReducedFit.create(rf_phot)
redf_phot.build_file()

# %% [markdown]

# As we can see, the astrometric calibration was done with the blind astrometry.net
# solver as before.

# Next, if we go to the polarimetry field, we see that there are not enough stars in
# the image to do the blind astrometric calibration.

# %% [markdown]
# The calibration of this file will use the previously calibrated photometry field
# and compare quads of stars in both images to find the transformation between
# them, as we can see in the log:

# %%
redf_pol = ReducedFit.create(rf_pol)
redf_pol.build_file()

# %% 
# !ls {redf_pol.filedpropdir}

from IPython.display import Image
Image(filename=f"{redf_pol.filedpropdir}/astrometry_matched_quads.png")

# %% [markdown]
# The image shows the quads of stars that were matched in both images, their
# distance in the hash space and the distance in the image space.

# This quad matching is necessary because the number of stars in the polarimetry
# field is not enough for the astrometry.net solver to work, plus the scale of the
# images is too small for the default index files. This quad matching needs at 
# least 4 (ideally a few more) stars in the image, and a previously calibrated 
# photometry field. If any of these conditions are not met, other methods must
# be used to calibrate the image (if there are only two bright sources at the 
# right distance, can be assumed that they are the target star, and if there are
# between 3 and 4-6 stars, we can try to match to known sources in the field).

# The process is implemented in the DIPOL._build_wcs_for_polarimetry_images_photo_quads
# method, which is called by the ReducedFit.astrometric_calibration method.
# We could also manually follow the process:

# %%
import numpy as np
import itertools
from iop4lib.instruments import DIPOL
from iop4lib.utils.astrometry import BuildWCSResult
from photutils.aperture import CircularAperture
from pathlib import Path
import matplotlib as mplt
import matplotlib.pyplot as plt

# get the subframe of the photometry field that corresponds to this polarimetry field, (approx)
x_start = redf_pol.rawfit.header['XORGSUBF']
y_start = redf_pol.rawfit.header['YORGSUBF']

x_end = x_start + redf_pol.rawfit.header['NAXIS1']
y_end = y_start + redf_pol.rawfit.header['NAXIS2']

idx = np.s_[y_start:y_end, x_start:x_end]

photdata_subframe = redf_phot.mdata[idx]

# find 10 brightest sources in each field

sets_L = list()

for redf, data in zip([redf_pol, redf_phot], [redf_pol.mdata, photdata_subframe]):

    positions = DIPOL._estimate_positions_from_segments(redf=redf, data=data, n_seg_threshold=1.5, npixels=32, centering=None, fwhm=1.0)
    positions = positions[:10]

    sets_L.append(positions)

fig = plt.figure(figsize=(12,6), dpi=iop4conf.mplt_default_dpi)
axs = fig.subplots(nrows=1, ncols=2)

for ax, data, positions in zip(axs, [redf_pol.mdata, photdata_subframe], sets_L):
    imshow_w_sources(data, pos1=positions, ax=ax)
    candidates_aps = CircularAperture(positions[:2], r=10.0)
    candidates_aps.plot(ax, color="b")
    for i, (x,y) in enumerate(positions):
        ax.text(x, y, f"{i}", color="orange", fontdict={"size":14, "weight":"bold"})#, verticalalignment="center", horizontalalignment="center") 
    ax.plot([data.shape[1]//2], [data.shape[0]//2], '+', color='y', markersize=10)
    
axs[0].set_title("Polarimetry field")
axs[1].set_title("Photometry field")
fig.suptitle("astrometry detected_sources")
plt.show()

# %%

# Build the quads for each field
quads_1 = np.array(list(itertools.combinations(sets_L[0], 4)))
quads_2 = np.array(list(itertools.combinations(sets_L[1], 4)))

from iop4lib.utils.quadmatching import hash_ish, distance, order, qorder_ish, find_linear_transformation
hash_func, qorder = hash_ish, qorder_ish

hashes_1 = np.array([hash_func(quad) for quad in quads_1])
hashes_2 = np.array([hash_func(quad) for quad in quads_2])

all_indices = np.array(list(itertools.product(range(len(quads_1)),range(len(quads_2)))))
all_distances = np.array([distance(hashes_1[i], hashes_2[j]) for i,j in all_indices])

idx = np.argsort(all_distances)
all_indices = all_indices[idx]
all_distances = all_distances[idx]

indices_selected = all_indices[:6]
colors = ["r", "g", "b", "c", "m", "y"]

# %%

figs, axs = zip(*[plt.subplots(figsize=(6,4), dpi=iop4conf.mplt_default_dpi) for _ in range(2)])

for (i,j), color in list(zip(indices_selected, colors)): 
    
    tij = find_linear_transformation(qorder(quads_1[i]), qorder(quads_2[j]))[1]

    for ax, fig, data, quad, positions in zip(axs, figs, [redf_pol.mdata, photdata_subframe], [quads_1[i], quads_2[j]], sets_L):
        imshow_w_sources(data, pos1=positions, ax=ax)
        x, y = np.array(order(quad)).T
        ax.fill(x, y, edgecolor='k', fill=True, facecolor=mplt.colors.to_rgba(color, alpha=0.2))

plt.show()


from ..enums import *
from .epoch import Epoch
from .rawfit import RawFit
from .astrosource import AstroSource
from .masterbias import MasterBias
from .masterflat import MasterFlat
from .masterdark import MasterDark
from .reducedfit import ReducedFit
from .photopolresult import PhotoPolResult, PhotoPolResultReducedFitRelation
from .aperphotresult import AperPhotResult


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports

# other imports
import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
import astrometry
import numpy as np
import math

# iop4lib imports
from iop4lib.enums import *
from .instrument import Instrument
from iop4lib.telescopes import CAHAT220

# logging
import logging
logger = logging.getLogger(__name__)

import typing
if typing.TYPE_CHECKING:
    from iop4lib.db import RawFit, ReducedFit, Epoch

class CAFOS(Instrument):
        
    name = "CAFOS2.2"
    telescope = CAHAT220.name

    instrument_kw_L = ["CAFOS 2.2"]

    arcsec_per_pix = 0.530
    gain_e_adu = 1.45
    field_width_arcmin = 34.0
    field_height_arcmin = 34.0

    required_masters = ['masterbias', 'masterflat']

    # pre computed pairs distances to use in the astrometric calibrations
    # obtained from calibrated fields
    
    # computed with:
    # > In [1]: qs = ReducedFit.objects.filter(flags__has=ReducedFit.FLAGS.BUILT_REDUCED, instrument="CAFOS2.2").all()
    # > In [2]: disp_sign_mean = np.mean([redf.astrometry_info[-1]['seg_disp_sign'] for redf in qs[len(qs)-300:len(qs)-1]], axis=0)
    # > In [3]: disp_sign_std = np.std([redf.astrometry_info[-1]['seg_disp_sign'] for redf in qs[len(qs)-300:len(qs)-1]], axis=0)

    disp_sign_mean, disp_sign_std = np.array([-35.72492116, -0.19719535]), np.array([1.34389, 1.01621491])
    disp_mean, disp_std = np.abs(disp_sign_mean), disp_sign_std


    @classmethod
    def classify_juliandate_rawfit(cls, rawfit):
        """
        CAHA T220 has a DATE keyword in the header in ISO format.
        """
        import astropy.io.fits as fits
        from astropy.time import Time 
        date = fits.getheader(rawfit.filepath, ext=0)["DATE"]
        jd = Time(date, format='isot', scale='utc').jd
        rawfit.juliandate = jd

    @classmethod
    def classify_imgtype_rawfit(cls, rawfit):
        """
        CAHA T220 has a IMAGETYP keyword in the header: flat, bias, science
        """
        from iop4lib.db.rawfit import RawFit
        import astropy.io.fits as fits

        with fits.open(rawfit.filepath) as hdul:
            if hdul[0].header['IMAGETYP'] == 'flat':
                rawfit.imgtype = IMGTYPES.FLAT
            elif hdul[0].header['IMAGETYP'] == 'bias':
                rawfit.imgtype = IMGTYPES.BIAS
            elif hdul[0].header['IMAGETYP'] == 'science':
                rawfit.imgtype = IMGTYPES.LIGHT
            elif hdul[0].header['IMAGETYP'] == 'dark':
                rawfit.imgtype = IMGTYPES.DARK
            else:
                logger.error(f"Unknown image type for {rawfit.fileloc}.")
                rawfit.imgtype = IMGTYPES.ERROR
                raise ValueError
            
    @classmethod
    def classify_band_rawfit(cls, rawfit):
        """
        Older data (e.g. 2007): INSFLNAM = 'John R' or INSFLNAM = 'Cous R'
        New data (e.g. 2022): INSFLNAM = 'BessellR'

        There are also images in the archive with INSFLNAM = 'John V' and INSFLNAM = 'John I', and INSFLNAM = 'free'
        """

        from iop4lib.db.rawfit import RawFit
        import astropy.io.fits as fits

        if 'INSFLNAM' in rawfit.header:
            if any([rawfit.header['INSFLNAM'] == kw for kw in ['BessellR', 'John R', 'Cous R', 'CousinsR', 'JohnsonR']]):
                rawfit.band = BANDS.R
            elif any([rawfit.header['INSFLNAM'] == kw for kw in ['BessellI', 'John I', 'Cous I', 'CousinsI', 'JohnsonI']]):
                rawfit.band = BANDS.I
            elif any([rawfit.header['INSFLNAM'] == kw for kw in ['BessellV', 'John V', 'Cous V', 'CousinsV', 'JohnsonV']]):
                rawfit.band = BANDS.V
            elif any([rawfit.header['INSFLNAM'] == kw for kw in ['BessellB', 'John B', 'Cous B', 'CousinsB', 'JohnsonB']]):
                rawfit.band = BANDS.B
            elif any([rawfit.header['INSFLNAM'] == kw for kw in ['BessellU', 'John U', 'Cous U', 'CousinsU', 'JohnsonU']]):
                rawfit.band = BANDS.U
            else:
                logger.error(f"{rawfit}: unknown filter {rawfit.header['INSFLNAM']}.")
                rawfit.band = BANDS.ERROR
                raise ValueError(f"{rawfit}: unknown filter {rawfit.header['INSFLNAM']}.")
        else: 
            rawfit.band = BANDS.ERROR
            raise ValueError(f"{rawfit}: INSFLNAM keyword not present.")

    @classmethod
    def classify_obsmode_rawfit(cls, rawfit):
        """
        For CAHA T220, if we are dealing with polarimetry, we have:
        INSTRMOD:	Polarizer
        INSPOFPI 	Wollaston
        INSPOROT 	0.0, 22.48, 67.48
        
        I HAVE NOT FOUND YET OTHER VALUES THAT ARE NOT THIS, PRINT A WARNING OTHERWISE.
        """
        from iop4lib.db.rawfit import RawFit

        if rawfit.header['INSTRMOD'] == 'Polarizer' and rawfit.header['INSPOFPI'] == 'Wollaston':
            rawfit.obsmode = OBSMODES.POLARIMETRY
            rawfit.rotangle = float(rawfit.header['INSPOROT'])

            if rawfit.imgtype == IMGTYPES.BIAS:
                logger.debug(f"Probably not important, but {rawfit.fileloc} is BIAS but has polarimetry keywords, does it makes sense?")
        elif rawfit.header['INSTRMOD'] == 'Polarizer' and rawfit.header['INSPOFPI'] == 'FREE':
            rawfit.obsmode = OBSMODES.PHOTOMETRY
        else:
            logger.error("Not implemented, please check the code.")

    @classmethod
    def get_header_hintcoord(cls, rawfit):
        """ Get the position hint from the FITS header as a coordinate.

        Images from CAFOS T2.2 have RA, DEC in the header, both in degrees.
        """

        hint_coord = SkyCoord(Angle(rawfit.header['RA'], unit=u.deg), Angle(rawfit.header['DEC'], unit=u.deg), frame='icrs')
        return hint_coord
    
    @classmethod
    def get_astrometry_position_hint(cls, rawfit, allsky=False, n_field_width=1.5, hintsep=None):
        """ Get the position hint from the FITS header as an astrometry.PositionHint object. 

        Parameters
        ----------
            allsky: bool, optional
                If True, the hint will cover the whole sky, and n_field_width and hintsep will be ignored.
            n_field_width: float, optional
                The search radius in units of field width. Default is 1.5.
            hintsep: Quantity, optional
                The search radius in units of degrees.
        """        

        hintcoord = cls.get_header_hintcoord(rawfit)

        if allsky:
            hintsep = 180.0 * u.deg
        else:
            if hintsep is None:
                hintsep = n_field_width * u.Quantity("16 arcmin") # 16 arcmin is the full field size of the CAFOS T2.2, our cut is smaller (6.25, 800x800, but the pointing kws might be from anywhere in the full field)

        return astrometry.PositionHint(ra_deg=hintcoord.ra.deg, dec_deg=hintcoord.dec.deg, radius_deg=hintsep.to_value(u.deg))
    
    @classmethod
    def get_astrometry_size_hint(cls, rawfit):
        """ Get the size hint for this telescope / rawfit.

        from http://w3.caha.es/CAHA/Instruments/CAFOS/cafos22.html
        pixel size in arcmin is around : ~0.530 arcsec
        field size (diameter is) 16.0 arcmin (for 2048 pixels)
        it seems that this is for 2048x2048 images, our images are 800x800 but the fitsheader DATASEC
        indicates it is a cut
        """
        
        return astrometry.SizeHint(lower_arcsec_per_pixel=0.95*cls.arcsec_per_pix,  upper_arcsec_per_pixel=1.05*cls.arcsec_per_pix)

    @classmethod
    def has_pairs(cls, fit_instance):
        """ At the moment, CAFOS polarimetry. """
        return (fit_instance.obsmode == OBSMODES.POLARIMETRY)

    @classmethod
    def compute_relative_polarimetry(cls, polarimetry_group):
        """ Computes the relative polarimetry for a polarimetry group for CAFOS observations.
        
        .. note::
            CAFOS Polarimetry observations are done with a system consisting of a half-wave plate (HW) and a Wollaston prism (P).

            The rotation angle theta_i refers to the angle theta_i between the HW plate and its fast (extraordinary) axes.

            The effect of the HW is to rotate the polarization vector by 2*theta_i, and the effect of the Wollaston prism is to split 
            the beam into two beams polarized in orthogonal directions (ordinary and extraordinary).

            An input polarized beam with direction v will be rotated by HW by 2*theta_i. The O and E fluxes will be the projections of the
            rotated vector onto the ordinary and extraordinary directions of the Wollaston prism (in absolute values since -45 and 45 
            polarization directions are equivalent). A way to write this is:

            fo(theta_i) = abs( <HW(theta_i)v,e_i> ) = abs ( <R(2*theta_i)v,e_i> ), where <,> denotes the scalar product and R is the rotation matrix.

            Therefore the following observed fluxes should be the same (ommiting the abs for clarity):

            fo(0º) = <v,e_1> = <v,R(-90)R(+90)e_1> = <R(90),R(90)e_i> = <HW(45),R(90)e_1> = fe(45º)
            fo(22º) = <HW(22)v,e_1> = <R(45)v,e_1> = <R(90)R(45)v,R(90)e_1> = <R(135)v,e_1> = <HW(67),R(90)e_1> = fe(67º)
            fo(45º) = <HW(45)v,e_1> = <R(90)v,e_1> = <v,R(-90)e_1> = -<v,e_2> = fe(0º)
            fo(67º) = <HW(67)v,e_1> = <R(135)v,e_1> = <R(90)R(45)v,e_1> = <R(45)v,R(-90)e_1> = <HW(22),R(-90)e_1> = fe(22º)

            See https://arxiv.org/pdf/astro-ph/0509153 (doi 10.1086/497581) for the formulas relating these fluxes to 
            the Stokes parameters.

        .. note::
            This rotation angle has a different meaning than for OSN-T090 Polarimetry observations. For them, it is the rotation angle of a polarized filter
            with respect to some reference direction. Therefore we have the equivalencies (again ommiting the abs for clarity):
            
            OSN(45º) = <v,R(45)e_1> = <R(45)v,R(45)R(45)e_1> = <HW(22),R(90)e_1> = fE(22º) = fO(67º)
            OSN(90º) = <v,R(90)e_1> = <R(90)v,R(90)R(90)e_1> = fO(45º)
            OSN(-45º) = OSN(135º) = abs(<v,R(-45)e_1>) = <R(45)v,e_1> = <R(135)v,R(90)e_1> = fE(67º) = fO(22º)
            OSN(0º) = <v,e_1> = <v,e_1> = fO(0º)
        """
        
        from iop4lib.db.aperphotresult import AperPhotResult
        from iop4lib.db.photopolresult import PhotoPolResult

        # Perform some checks on the group

        ## get the band of the group

        bands = [reducedfit.band for reducedfit in polarimetry_group]

        if len(set(bands)) == 1:
            band = bands[0]
        else: # should not happen
            raise Exception(f"Can not compute relative polarimetry for a group with different bands: {bands}")

        ## check obsmodes

        if not all([reducedfit.obsmode == OBSMODES.POLARIMETRY for reducedfit in polarimetry_group]):
            raise Exception(f"This method is only for polarimetry images.")
        
        ## check sources in the fields

        sources_in_field_qs_list = [reducedfit.sources_in_field.all() for reducedfit in polarimetry_group]
        group_sources = set.intersection(*map(set, sources_in_field_qs_list))

        if len(group_sources) == 0:
            logger.error("No common sources in field for all polarimetry groups.")
            return
        
        if group_sources != set.union(*map(set, sources_in_field_qs_list)):
            logger.warning(f"Sources in field do not match for all polarimetry groups: {set.difference(*map(set, sources_in_field_qs_list))}")

        ## check rotation angles

        rot_angles_available = set([redf.rotangle for redf in polarimetry_group])
        rot_angles_required = {0.0, 22.48, 44.98, 67.48}

        if not rot_angles_available.issubset(rot_angles_required):
            logger.warning(f"Rotation angles missing: {rot_angles_required - rot_angles_available}")

        # 1. Compute all aperture photometries

        aperpix, r_in, r_out, fit_res_dict = cls.estimate_common_apertures(polarimetry_group, reductionmethod=REDUCTIONMETHODS.RELPHOT)
        target_fwhm = fit_res_dict['mean_fwhm']
        
        logger.debug(f"Computing aperture photometries for the {len(polarimetry_group)} reducedfits in the group with target aperpix {aperpix:.1f}.")

        for reducedfit in polarimetry_group:
            cls.compute_aperture_photometry(reducedfit, aperpix, r_in, r_out)

        # 2. Compute relative polarimetry for each source (uses the computed aperture photometries)

        logger.debug("Computing relative polarimetry.")

        photopolresult_L = list()

        for astrosource in group_sources:
            logger.debug(f"Computing relative polarimetry for {astrosource}.")

            # if any angle is missing for some pair, it uses the equivalent angle of the other pair

            qs = AperPhotResult.objects.filter(reducedfit__in=polarimetry_group, astrosource=astrosource, aperpix=aperpix, flux_counts__isnull=False)

            if not qs.exists():
                logger.error(f"No aperture photometry results found for {astrosource}, skipping.")
                continue

            equivs = ((('O',0.0),   ('E',44.98)),
                      (('O',22.48), ('E',67.48)),
                      (('O',44.98), ('E',0.0)),
                      (('O',67.48), ('E',22.48)),
                      (('E',0.0),   ('O',44.98)),
                      (('E',22.48), ('O',67.48)),
                      (('E',44.98),  ('O',0.0)),
                      (('E',67.48), ('O',22.48)))

            flux_D = dict()
            _skip = False
            for equiv in equivs:
                if qs.filter(pairs=equiv[0][0], reducedfit__rotangle=equiv[0][1]).exists():
                    flux_D[(equiv[0][0], equiv[0][1])] = qs.filter(pairs=equiv[0][0], reducedfit__rotangle=equiv[0][1]).values_list("flux_counts", "flux_counts_err").last()
                elif qs.filter(pairs=equiv[1][0], reducedfit__rotangle=equiv[1][1]).exists():
                    logger.warning(f"Missing flux for {astrosource} {equiv[0][0]} {equiv[0][1]}, using {equiv[1][0]} {equiv[1][1]}")
                    flux_D[(equiv[0][0], equiv[0][1])] = qs.filter(pairs=equiv[1][0], reducedfit__rotangle=equiv[1][1]).values_list("flux_counts", "flux_counts_err").last()
                else:
                    logger.error(f"Missing flux for {astrosource} {equiv[0][0]} {equiv[0][1]} and {equiv[1][0]} {equiv[1][1]}, skipping this source.")
                    _skip = True
                    break
            if _skip:
                continue
            
            flux_O_0, flux_O_0_err = flux_D[('O',0.0)]
            flux_O_22, flux_O_22_err = flux_D[('O',22.48)]
            flux_O_45, flux_O_45_err = flux_D[('O',44.98)]
            flux_O_67, flux_O_67_err = flux_D[('O',67.48)]
            flux_E_0, flux_E_0_err = flux_D[('E',0.0)]
            flux_E_22, flux_E_22_err = flux_D[('E',22.48)]
            flux_E_45, flux_E_45_err = flux_D[('E',44.98)]
            flux_E_67, flux_E_67_err = flux_D[('E',67.48)]

            fluxes_O = np.array([flux_O_0, flux_O_22, flux_O_45, flux_O_67])
            fluxes_E = np.array([flux_E_0, flux_E_22, flux_E_45, flux_E_67])

            # logger.debug(f"Fluxes_O: {fluxes_O}")
            # logger.debug(f"Fluxes_E: {fluxes_E}")

            fluxes = (fluxes_O + fluxes_E) / 2.
            flux_mean = fluxes.mean()
            flux_err = fluxes.std() / math.sqrt(len(fluxes))

            RQ = np.sqrt((flux_O_0 / flux_E_0) / (flux_O_45 / flux_E_45))
            dRQ = RQ / 2 * math.sqrt((flux_O_0_err / flux_O_0) ** 2 + (flux_E_0_err / flux_E_0) ** 2 + (flux_O_45_err / flux_O_45) ** 2 + (flux_E_45_err / flux_E_45) ** 2)

            RU = np.sqrt((flux_O_22 / flux_E_22) / (flux_O_67 / flux_E_67))
            dRU = RU / 2 * math.sqrt((flux_O_22_err / flux_O_22) ** 2 + (flux_E_22_err / flux_E_22) ** 2 + (flux_O_67_err / flux_O_67) ** 2 + (flux_E_67_err / flux_E_67) ** 2)

            Q_I = (RQ - 1) / (RQ + 1)
            dQ_I = math.fabs( RQ / (RQ + 1) ** 2 * dRQ)
            U_I = (RU - 1) / (RU + 1)
            dU_I = math.fabs( RU / (RU + 1) ** 2 * dRU)

            P = math.sqrt(Q_I ** 2 + U_I ** 2)
            dP = 1/P * math.sqrt(Q_I**2 * dQ_I**2 + U_I**2 * dU_I**2)

            Theta_0 = 0
        
            if Q_I >= 0:
                Theta_0 = math.pi 
                if U_I > 0:
                    Theta_0 = -1 * math.pi
                # if Q_I < 0:
                #     Theta_0 = math.pi / 2
                
            Theta = 0.5 * math.degrees(math.atan(U_I / Q_I) + Theta_0)
            dTheta = 0.5 * 180.0 / math.pi * (1 / (1 + (U_I/Q_I) ** 2)) * math.sqrt( (dU_I/Q_I)**2 + (U_I*dQ_I/Q_I**2)**2 )

            # compute instrumental magnitude

            if flux_mean <= 0.0:
                logger.warning(f"{polarimetry_group=}: negative flux mean encountered while relative polarimetry for {astrosource=} ??!! It will be nan, but maybe we should look into this...")

            mag_inst = -2.5 * np.log10(flux_mean) # slower than math.log10 but returns nan when flux < 0 instead of throwing error (see https://github.com/juanep97/iop4/issues/24)
            mag_inst_err = math.fabs(2.5 / math.log(10) * flux_err / flux_mean)

            # if the source is a calibrator, compute also the zero point

            if astrosource.is_calibrator:
                mag_known = getattr(astrosource, f"mag_{band}")
                mag_known_err = getattr(astrosource, f"mag_{band}_err", None) or 0.0

                if mag_known is None:
                    logger.warning(f"Calibrator {astrosource} has no magnitude for band {band}.")
                    mag_zp = np.nan
                    mag_zp_err = np.nan
                else:
                    mag_zp = mag_known - mag_inst
                    # mag_zp_err = math.sqrt(mag_known_err ** 2 + mag_inst_err ** 2)
                    mag_zp_err = math.fabs(mag_inst_err) # do not add error on literature magnitude
            else:
                mag_zp = None
                mag_zp_err = None

            # save the results
                    
            result = PhotoPolResult.create(reducedfits=polarimetry_group, 
                                           astrosource=astrosource, 
                                           reduction=REDUCTIONMETHODS.RELPOL, 
                                           mag_inst=mag_inst, mag_inst_err=mag_inst_err, 
                                           mag_zp=mag_zp, mag_zp_err=mag_zp_err,
                                           flux_counts=flux_mean, 
                                           p=P, p_err=dP, 
                                           chi=Theta, chi_err=dTheta,
                                           aperpix=aperpix)
            
            result.aperphotresults.set(qs, clear=True)
            
            photopolresult_L.append(result)

        if not photopolresult_L:
            logger.error("No results could be computed for this group.")
            return
        
        # 3. Get average zero point from zp of all calibrators in the group

        calib_mag_zp_array = np.array([result.mag_zp or np.nan for result in photopolresult_L if result.astrosource.is_calibrator]) # else it fills with None also and the dtype becomes object
        calib_mag_zp_array = calib_mag_zp_array[~np.isnan(calib_mag_zp_array)]

        calib_mag_zp_array_err = np.array([result.mag_zp_err or np.nan for result in photopolresult_L if result.astrosource.is_calibrator])
        calib_mag_zp_array_err = calib_mag_zp_array_err[~np.isnan(calib_mag_zp_array_err)]

        if len(calib_mag_zp_array) == 0:
            logger.error(f"Can not compute magnitude during relative photo-polarimetry without any calibrators for this reduced fit.")

        zp_avg = np.nanmean(calib_mag_zp_array)
        zp_std = np.nanstd(calib_mag_zp_array)

        zp_err = np.sqrt(np.nansum(calib_mag_zp_array_err ** 2)) / len(calib_mag_zp_array_err)
        zp_err = math.sqrt(zp_err ** 2 + zp_std ** 2)

        # 4. Compute the calibrated magnitudes for non-calibrators in the group using the averaged zero point

        for result in photopolresult_L:

            if result.astrosource.is_calibrator:
                continue

            result.mag_zp = zp_avg
            result.mag_zp_err = zp_err
        
            result.mag = result.mag_inst + zp_avg
            result.mag_err = math.sqrt(result.mag_inst_err ** 2 + zp_err ** 2)

            result.save()

        # 5. Save results
        for result in photopolresult_L:
            result.save()



    @classmethod
    def _build_shotgun_params(cls, redf: 'ReducedFit'):
        shotgun_params_kwargs = dict()

        shotgun_params_kwargs["d_eps"] = [1.0] #[1*np.linalg.norm(cls.disp_std)]
        shotgun_params_kwargs["dx_eps"] = [1.0] #[1*cls.disp_std[0]]
        shotgun_params_kwargs["dy_eps"] = [1.0] #[1*cls.disp_std[1]]
        shotgun_params_kwargs["dx_min"] = [(cls.disp_mean[0] - 5*cls.disp_std[0])]
        shotgun_params_kwargs["dx_max"] = [(cls.disp_mean[0] + 5*cls.disp_std[0])]
        shotgun_params_kwargs["dy_min"] = [(cls.disp_mean[1] - 5*cls.disp_std[1])]
        shotgun_params_kwargs["dy_max"] = [(cls.disp_mean[1] + 5*cls.disp_std[1])]
        shotgun_params_kwargs["d_min"] = [np.linalg.norm(cls.disp_mean) - 3*np.linalg.norm(cls.disp_std)]
        shotgun_params_kwargs["d_max"] = [np.linalg.norm(cls.disp_mean) + 3*np.linalg.norm(cls.disp_std)]
        shotgun_params_kwargs["bins"] = [400]
        shotgun_params_kwargs["hist_range"] = [(0,400)]

        if redf.header_hintobject is not None and redf.header_hintobject.name == "1101+384":
            shotgun_params_kwargs["bkg_filter_size"] = [3]
            shotgun_params_kwargs["bkg_box_size"] = [16]
            shotgun_params_kwargs["seg_fwhm"] = [1.0]
            shotgun_params_kwargs["npixels"] = [8, 16]
            shotgun_params_kwargs["n_rms_seg"] = [3.0, 1.5, 1.2, 1.1, 1.0]


        return shotgun_params_kwargs

    @classmethod
    def build_wcs(cls, reducedfit: 'ReducedFit', summary_kwargs : dict = None):
        return super().build_wcs(reducedfit, shotgun_params_kwargs=cls._build_shotgun_params(reducedfit), summary_kwargs=summary_kwargs)

    @classmethod
    def estimate_common_apertures(cls, reducedfits, reductionmethod=None, fit_boxsize=None, search_boxsize=(30,30), fwhm_min=2, fwhm_max=20):
        return super().estimate_common_apertures(reducedfits, reductionmethod=reductionmethod, fit_boxsize=fit_boxsize, search_boxsize=search_boxsize, fwhm_min=fwhm_min, fwhm_max=fwhm_max)


import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

from django.db import models

import os
import datetime
import numpy as np
import astropy.io.fits as fits
import datetime

from ..enums import *
from .fitfilemodel import FitFileModel
from .fields import FlagChoices, FlagBitField

import logging
logger = logging.getLogger(__name__)

class MasterBias(FitFileModel):
    """A class representing a master bias for an epoch."""

    margs_kwL = ['epoch', 'instrument', 'imgsize']

    imgtype = IMGTYPES.BIAS

    # Database fields
    rawfits = models.ManyToManyField('RawFit', related_name='built_masterbias')

    # flags

    class FLAGS(FlagChoices):
        IGNORE = 1 << 1
        
    flags = FlagBitField(choices=FLAGS.choices(), default=0)

    # fields corresponding to MasterBias kw arguments (margs_kwL)
    
    epoch = models.ForeignKey('Epoch', on_delete=models.CASCADE, related_name='masterbias') 
    instrument = models.CharField(max_length=20, choices=INSTRUMENTS.choices)
    imgsize = models.CharField(max_length=12, null=True)

    class Meta:
        app_label = 'iop4api'
        verbose_name = "Master Bias"
        verbose_name_plural = "Master Biases"

    # Properties

    @property
    def filename(self):
        return f"masterbias_id{self.id}.fits"
    
    @property
    def filepath(self):
        return os.path.join(self.epoch.masterbiasdir, self.filename)

    @property
    def fileloc(self):
        return f"{self.epoch.epochname}/{self.filename}"
    
    # Filed Properties location

    @property
    def filedpropdir(self):
       return os.path.join(self.epoch.calibrationdir, 
                            self.filename +  '.d', 
                            self.__class__.__name__)
        
    @property
    def margs(self):
        """Return a dict of the arguments used to build this MasterFlat.
        """

        return {'epoch':self.epoch, 'instrument':self.instrument, 'imgsize':self.imgsize}
    
    # repr and str

    @classmethod
    def margs2str(cls, margs):
       """Class method to build a nice string rep of the arguments of a MasterFlat.
       """

       return f"{margs['epoch'].epochname} | {margs['instrument']} | {margs['imgsize']}"

    def __repr__(self):
        return f"MasterBias.objects.get(id={self.id!r})"
    
    def __str__(self):
        return f"<MasterBias {self.id!r} | {MasterBias.margs2str(self.margs)}>"
    
    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text(f"{self!r}")
        else:
            with p.group(4, f'<{self.__class__.__name__}(', ')>'):
                p.text(f"id: {self.id},")
                for k,v in self.margs.items():
                    p.breakable()
                    p.text(f"{k}: {v}")
                

    
    # Constructor

    @classmethod
    def create(cls,
               rawfits=None,
               auto_merge_to_db=True,
               force_rebuild=False,
               **kwargs):
        """Create a MasterBias object for the given epoch, reusing an existing DB entry if it exists.

        It will overwrite any existing MasterBias for that epoch, or create a new one if it doesn't exist.
    
        Parameters
        ----------
        rawfits : list of RawFit, optional
            List of RawFit objects used to build the MasterBias.
        auto_merge_to_db : bool, optional
            If True, automatically merge the created MasterBias object to the database.
        force_rebuild : bool, optional
            If True, force rebuilding the MasterBias even if it already exists.

        Other Parameters
        ----------------
        `**kwargs` :
            Keyword arguments passed to the constructor.

        """

        margs = {k:kwargs[k] for k in MasterBias.margs_kwL if k in kwargs}

        from iop4lib.db import RawFit

        if (mb := MasterBias.objects.filter(**margs).first()) is not None:
            logger.debug(f"DB entry for {mb} already exists, using it instead.")
        else:
            logger.debug(f"A DB entry for MasterBias {MasterBias.margs2str(margs)} will be created.")
            mb = cls(**margs)
            mb.save()

        if rawfits is None:
            rawfits = RawFit.objects.filter(imgtype=IMGTYPES.BIAS, **margs)
            logger.debug(f"Found {len(rawfits)} bias raw files for {mb}.")
        else:
            logger.debug(f"Using {len(rawfits)} bias raw files for {mb}.")

        mb.rawfits.set(rawfits)

        # Build the file
        if not mb.fileexists or force_rebuild:
            logger.info(f"Building file")
            try:
                mb.build_file()
            except Exception as e:
                logger.error("An error occurred while building the masterbias, deleting it and raising Exception")
                mb.delete()
                raise Exception(f"An error occurred while building the MasterBias for {margs}: {e}")
            else:
                logger.info("Built masterbias successfully")

        # merge to DB

        if auto_merge_to_db:
            mb.save()

        return mb

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.auto_merge_to_db = True
    
    
    # Methods

    def build_file(self):

        logger.debug(f"Getting data from files")

        if self.rawfits.count() == 0:
            logger.error(f"No rawfits for {self}")
            raise Exception(f"No rawfits for {self}")
        
        data_L = []
        for rf in self.rawfits.all():
            with fits.open(rf.filepath) as hdul:
                data_L.append(hdul[0].data)

        logger.debug(f"Computing median")

        data = np.nanmedian(data_L, axis=0)

        logger.debug(f"Building header")

        header = fits.Header()
        header['TELESCOP'] = self.epoch.telescope
        header['NIGHT'] = self.epoch.yyyymmdd
        header['EPOCH'] = self.epoch.epochname
        header['IMGSIZE'] = self.imgsize
        header['IMGTYPE'] = 'masterbias'
        header['DATECREA'] = datetime.datetime.now(datetime.UTC).isoformat(timespec="milliseconds")
        header['NRAWFITS'] = self.rawfits.count()

        logger.debug(f"Building HDU")

        if not os.path.isdir(self.epoch.masterbiasdir):
            logger.debug(f"Creating directory {self.epoch.masterbiasdir}")
            os.makedirs(self.epoch.masterbiasdir)

        hdu = fits.PrimaryHDU(data, header=header)

        logger.debug(f"Writing MasterBias to {self.filepath}")
        hdu.writeto(self.filepath, overwrite=True)


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports

# other imports
from abc import ABCMeta, abstractmethod
import os
import re
import ftplib
from pathlib import Path
import astropy.io.fits as fits
import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
import astrometry
import numpy as np
import math

# iop4lib imports
from iop4lib.enums import *
from .telescope import Telescope, FTPArchiveMixin, ReadOnlyClassProperty

# logging
import logging
logger = logging.getLogger(__name__)

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from iop4lib.db import RawFit, ReducedFit, Epoch

class CAHAT220(FTPArchiveMixin, Telescope, metaclass=ABCMeta):
    """
    CAHA T220 telescope.

    CAHA has a per-program ftp login (PI user and password in config).
    The file structure when you login is a list of files as:
    {yymmdd}_CAFOS/
    where CAFOS refers to the polarimeter
    and inside each folder there are the files for that day.

    Currently only one instrument, CAFOS.
    """

    # telescope identification

    name = "CAHA-T220"
    abbrv = "T220"
    telescop_kw = "CA-2.2"

    # telescope specific properties

    # ftp connection details need to be properties so they can be overriden
    # otherwise they are defined at import time

    @ReadOnlyClassProperty
    def ftp_address(cls):
        return iop4conf.caha_address
    
    @ReadOnlyClassProperty
    def ftp_user(cls):
        return iop4conf.caha_user
    
    @ReadOnlyClassProperty
    def ftp_password(cls):
        return iop4conf.caha_password
    
    @ReadOnlyClassProperty
    def ftp_encoding(cls):
        return 'utf-8'

    re_expr_dirnames = re.compile(r"([0-9]{2}[0-9]{2}[0-9]{2})_CAFOS", flags=re.IGNORECASE)
    re_expr_fnames = re.compile(r".*\.fi?ts?", flags=re.IGNORECASE)

    # telescope specific methods

    # for the ftp archive mixin 

    @classmethod
    def remote_dirname_to_epochname(cls, dirname):
        matchs = cls.re_expr_dirnames.findall(dirname)
        if len(matchs) != 1:
            return None
        else:
            return f"{cls.name}/20{matchs[0][0:2]}-{matchs[0][2:4]}-{matchs[0][4:6]}"
        
    @classmethod
    def epochname_to_remote_dirname(cls, epochname):
        from iop4lib.db import Epoch
        tel, night = Epoch.epochname_to_tel_night(epochname)
        return night.strftime("%y%m%d_CAFOS")


# iop4lib config
import iop4lib
iop4conf = iop4lib.Config(config_db=False)

# django imports
from django.http import JsonResponse, HttpResponseBadRequest
from django.contrib.auth.decorators import permission_required
from django.views.decorators.csrf import csrf_exempt

# iop4lib imports
from iop4lib.db import PhotoPolResult
import json

# other imports

#logging
import logging
logger = logging.getLogger(__name__)


@csrf_exempt
@permission_required(["iop4api.change_photopolresult"])
def flag(request):

    try:
        payload = json.loads(request.body)
        flag = payload['flag']
        pks = payload['pk_array']
        vals = payload['vals']
    except:
        return HttpResponseBadRequest("Could not parse JSON")
    
    qs = PhotoPolResult.objects.filter(pk__in=pks)

    for r in qs:
        if vals[pks.index(r.pk)]:
            r.set_flag(flag)
        else:
            r.unset_flag(flag)

    PhotoPolResult.objects.bulk_update(qs, ['flags'])

    new_flag_dict = {k:v for k,v in PhotoPolResult.objects.filter(pk__in=pks).values_list('pk', 'flags')}

    logger.debug(f"{new_flag_dict=}")

    return JsonResponse({'success': True, 'flag': flag, 'new_flag_dict':new_flag_dict})




import pytest
from pathlib import Path

from .conftest import TEST_CONFIG

# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_path=TEST_CONFIG)

# other imports
import os
from pytest import approx

# logging
import logging
logger = logging.getLogger(__name__)

# fixtures
from .fixtures import load_test_catalog



@pytest.mark.django_db(transaction=True)
def test_build_multi_proc_photopol(load_test_catalog):
    """ Test the whole building process of reduced fits through multiprocessing 
    
    Also tests here relative photometry and polarimetry results and their 
    quality (value + uncertainties) (to avoud losing time reducing them
    in another test function).
    """

    from iop4lib.db import Epoch, RawFit, ReducedFit
    from iop4lib.enums import IMGTYPES, SRCTYPES, INSTRUMENTS

    # get epochs that have CAFOS observations in them
    from iop4lib.iop4 import list_local_epochnames
    epochname_L = list_local_epochnames()
    epoch_L = [Epoch.create(epochname=epochname) for epochname in epochname_L]
    epoch_L = [epoch for epoch in epoch_L if epoch.rawfits.filter(instrument=INSTRUMENTS.CAFOS).exists()]

    for epoch in epoch_L:
        epoch.build_master_biases()
        epoch.build_master_flats()

    # 1. Test multi-process reduced fits building

    iop4conf.nthreads = 4

    rawfits = RawFit.objects.filter(epoch__in=epoch_L, instrument=INSTRUMENTS.CAFOS, imgtype=IMGTYPES.LIGHT).all()
    
    Epoch.reduce_rawfits(rawfits)

    assert (ReducedFit.objects.filter(epoch__in=epoch_L).count() == len(rawfits))

    for redf in ReducedFit.objects.filter(epoch__in=epoch_L).all():
        assert (redf.has_flag(ReducedFit.FLAGS.BUILT_REDUCED))
        assert not (redf.has_flag(ReducedFit.FLAGS.ERROR_ASTROMETRY))

    from iop4lib.db import PhotoPolResult, AstroSource

    # 2. Test relative photo-polarimetry results

    epoch = Epoch.by_epochname("CAHA-T220/2022-09-18")

    epoch.compute_relative_polarimetry()

    qs_res = PhotoPolResult.objects.filter(epoch=epoch, astrosource__name="2200+420").all()

    # we expect only one photometry result target in this test dataset for this epoch and source
    assert qs_res.exclude(astrosource__in=AstroSource.objects.filter(is_calibrator=True)).count() == 1

    res = qs_res[0]

    # check that the result is correct compared to IOP3
    
    assert res.mag == approx(13.38, abs=max(1.5*res.mag_err, 0.05))
    assert res.mag_err < 0.08

    assert res.p == approx(10.9/100, abs=max(1.5*res.p_err, 1.0/100))
    assert res.p_err < 0.5

    assert res.chi == approx(25.2, abs=max(1.5*res.chi_err, 1.0))
    assert res.chi_err < 1.0



import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports
from django.db import models
from django.db.models import Exists, OuterRef
from django.db.models import Q, Avg

# iop4lib imports
from ..enums import *

# other imports
import os
import pypandoc 
import warnings
import astropy.io.fits as fits
from astropy.wcs import WCS
from astropy.coordinates import Angle, SkyCoord
import astropy.units as u
import math

# logging
import logging
logger = logging.getLogger(__name__)

class AstroSourceQuerySet(models.QuerySet):
    def with_is_calibrator(self):
        calibrator_subquery = self.filter(calibrators=OuterRef('pk'))
        return self.annotate(
            is_calibrator=Exists(calibrator_subquery)
        )

class AstroSourceManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    def get_queryset(self):
        return AstroSourceQuerySet(self.model).with_is_calibrator()
      
class AstroSource(models.Model):
    """ AstroSource model, representing an astronomical source in the IOP4 catalog.

    The coordinates (ra, dec) are interpreted as ICRS coordinates, following astropy `SkyCoord` default convention.
    """

    from iop4lib.enums import SRCTYPES

    # DB fields

    # Identification

    name = models.CharField(max_length=255, unique=True, help_text="Name of the source (must be unique)")

    # common to all sources

    other_names = models.CharField(max_length=255, null=True, blank=True, help_text="Alternative names for the source, separated by a semicolon ';'. It might be needed to correctly identify the target source of observations if observers used a different name.")
    ra_hms = models.CharField(max_length=255, help_text="Right ascension (ICRS) in hh:mm:ss format")
    dec_dms = models.CharField(max_length=255, help_text="Declination (ICRS) in dd:mm:ss format")
    srctype = models.CharField(max_length=255, choices=SRCTYPES.choices, help_text="Source type")  
    comment = models.TextField(null=True, blank=True, help_text="Any comment about the source (in Markdown format)")

    # Blazar fields

    redshift = models.FloatField(null=True, blank=True, help_text="Redshift of the source")
    
    # Calibration stars fields

    calibrates = models.ManyToManyField('AstroSource', related_name="calibrators", blank=True, help_text="sources that it calibrates (for calibrators only)")

    mag_R = models.FloatField(blank=True, null=True, help_text="Literature magnitude in R band (for calibrators only)")
    mag_R_err = models.FloatField(blank=True, null=True, help_text="Literature magnitude error in R band (for calibrators only)")

    mag_B = models.FloatField(blank=True, null=True, help_text="Literature magnitude in B band (for calibrators only)")
    mag_B_err = models.FloatField(blank=True, null=True, help_text="Literature magnitude error in B band (for calibrators only)")

    mag_V = models.FloatField(blank=True, null=True, help_text="Literature magnitude in V band (for calibrators only)")
    mag_V_err = models.FloatField(blank=True, null=True, help_text="Literature magnitude error in V band (for calibrators only)")

    mag_I = models.FloatField(blank=True, null=True, help_text="Literature magnitude in I band (for calibrators only)")
    mag_I_err = models.FloatField(blank=True, null=True, help_text="Literature magnitude error in I band (for calibrators only)")

    mag_U = models.FloatField(blank=True, null=True, help_text="Literature magnitude in U band (for calibrators only)")
    mag_U_err = models.FloatField(blank=True, null=True, help_text="Literature magnitude error in U band (for calibrators only)")


    # Natural key
    # allows us to relate the sources and calibration stars by names only

    # custom manager allows us to use natural keys when loading fixtures
    # and introduces virtual calibrator boolean field
    objects = AstroSourceManager()
    
    # this method allows us to dump using natural keys
    def natural_key(self):
            return (self.name,)
    
    class Meta:
        app_label = 'iop4api'
        verbose_name = "AstroSource"
        verbose_name_plural = "AstroSources"
    
    # repr and str

    def __repr__(self):
        return f"AstroSource.objects.get(name={self.name!r})"

    def __str__(self):
        return f"<AstroSource {self.name}>"
    
    def _repr_html_(self):
        lvl = logging.getLogger().level
        logging.getLogger().setLevel(logging.CRITICAL)
        res_html =  (f"{self.__class__.__name__}(name={self.name!r}):<br>\n"
                    f" - other_names: {self.other_names}<br>\n"
                    f" - ra={self.coord.ra:.4f}, dec={self.coord.dec:.4f}<br>\n"
                    f" - srctype: {self.srctype}<br>\n"
                    f" - comment: {self.comment_html}<br>\n")
        logging.getLogger().setLevel(lvl)
        return res_html
    
    # methods

    def is_in_field(self, wcs, height, width):
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                x, y = self.coord.to_pixel(wcs)
        except:
            return False
        else:
            if (0 <= x < height) and (0 <= y < width):
                return True
            else:
                return False
            
    # helper properties

    @property
    def other_names_list(self):
        if not self.other_names:
            return []
        elif ';' not in self.other_names:
            return [self.other_names]
        else:
            return [s.strip() for s in self.other_names.split(';')]
        
    @property
    def all_names_list(self):
        return [self.name] + self.other_names_list

    
    @property
    def coord(self):
        return SkyCoord(self.ra_hms, self.dec_dms, unit=(u.hourangle, u.deg))
    
    @property
    def comment_html(self):
        if '<br>' in self.comment:
             fmt = 'html'
             html_src = self.comment
        else:
             fmt = 'md'
             html_src = pypandoc.convert_text(self.comment, 'html', format=fmt, verify_format=False)

        return html_src

    @property
    def filedpropdir(self):
        return os.path.join(iop4conf.datadir, "astrosource", self.name)
    
    # Class methods

    @classmethod
    def get_catalog_as_dict(cls, *args, **kwargs):
        objs = cls.objects.filter(*args, **kwargs).all()
        catalog_dict = {obj.id: {'name': obj.name, 
                                 'other_names': obj.other_names,
                                 'srctype': obj.srctype,
                                 'comment': obj.comment,
                                 'coord': obj.coord,
                                 } for obj in objs}
        return catalog_dict
    
    @classmethod
    def get_sources_in_field(cls, wcs=None, width=None, height=None, fit=None, qs=None):
        r""" Get the sources in the field of view of the image.

        It accepts either a fit image or a wcs, height and width.
        If no query set is given, it will search the whole catalog,
        otherwise it will search the given query set.
        """

        if fit is not None:
            wcs = fit.wcs
            width, height = fit.width, fit.height

        if qs is None:
            qs = cls.objects.all()

        sources_in_field = list()

        import warnings
        for obj in qs:
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    x, y = obj.coord.to_pixel(wcs)
            except:
                pass
            else:
                if (0 <= x < height) and (0 <= y < width):
                    sources_in_field.append(obj)

        return sources_in_field


    @property
    def last_reducedfit(self):
        """Returns the last ReducedFit object associated with the source."""
        return self.in_reducedfits.order_by('-epoch__night').first()
    
    @property
    def last_night_mag_R(self):
        """Returns the average magnitude and error of the last night in the R band."""

        last_night = self.photopolresults.filter(band=BANDS.R).earliest('-epoch__night').epoch.night
        r_avg = self.photopolresults.filter(band=BANDS.R, epoch__night=last_night).aggregate(mag_avg=Avg('mag'), mag_err_avg=Avg('mag_err'))

        mag_r_avg = r_avg.get('mag_avg', None)
        mag_r_err_avg = r_avg.get('mag_err_avg', None)

        return mag_r_avg, mag_r_err_avg
    
    @property
    def texp_andor90(self):
        """Recommneded exposure time for Andor90, based on the last R magnitude and for a SNR of 150."""

        snr = 150
        last_night_mag_R, _ = self.last_night_mag_R

        if last_night_mag_R is None:
            return None

        texp = math.pow(snr,2) * 9.77 * 1e-16 * math.pow(10, 0.8*last_night_mag_R)

        if texp < 30:
            return 60
        elif texp <= 100:
            return 150
        elif texp <= 250:
            return 300
        elif texp <= 400:
            return 450
        elif texp <= 800:
            return 600
        else:
            return None
        
    @property
    def texp_andor150(self):
        """Recommneded exposure time for Andor150, based on the last night R magnitude and for a SNR of 150."""

        snr = 150
        last_night_mag_R, _ = self.last_night_mag_R

        if last_night_mag_R is None:
            return None
        
        texp = 0.36 * math.pow(snr,2) * 9.77 * 1e-16 * math.pow(10, 0.8*last_night_mag_R)

        if texp < 30:
            return 60
        elif texp <= 100:
            return 150
        elif texp <= 250:
            return 300
        elif texp <= 400:
            return 450
        else:
            return 600
    
    @property
    def texp_dipol(self):
        """Recommneded exposure time for DIPOL, based on the last night R magnitude and for a SNR of 150."""

        snr = 150
        last_night_mag_R, _ = self.last_night_mag_R

        if last_night_mag_R is None:
            return None
        
        texp = math.pow(snr,2) * 9.77 * 1e-16 * math.pow(10, 0.8*last_night_mag_R)

        if  texp <= 300:
            return math.ceil(texp / 10) * 10 + 10
        elif texp <= 2000:
            return 300
        else:
            return None
        
    @property
    def nreps_dipol(self):
        """Recommneded number of repetitions for DIPOL, based on the last night R magnitude and for a SNR of 150."""

        snr = 150
        last_night_mag_R, _ = self.last_night_mag_R

        if last_night_mag_R is None:
            return None
        
        texp = math.pow(snr,2) * 9.77 * 1e-16 * math.pow(10, 0.8*last_night_mag_R)

        if texp <= 20:
            return 8
        elif texp <= 40:
            return 4
        elif texp <= 80:
            return 2
        else:
            return 1


import pytest
from pathlib import Path

from .conftest import TEST_CONFIG

# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_path=TEST_CONFIG)

# other imports
import os
from pytest import approx

# logging
import logging
logger = logging.getLogger(__name__)

# fixtures
from .fixtures import load_test_catalog

@pytest.mark.django_db(transaction=True)
def test_epoch_creation(load_test_catalog):
    """ Test epoch creation """

    from iop4lib.db import Epoch

    assert (Epoch.objects.count() == 0)
    
    epoch = Epoch.create(epochname="OSN-T090/2022-09-08")

    assert (Epoch.objects.count() == 1)
    assert (epoch.rawfits.count() > 0)
    assert (epoch.rawfits.count() == len(os.listdir(Path(iop4conf.datadir) / "raw" / "OSN-T090" / "2022-09-08")))

@pytest.mark.django_db(transaction=True)
def test_epoch_masterbias_masterflats(load_test_catalog):
    """ Test masterbias and masterflats creation """
    from iop4lib.db import Epoch
    
    epoch = Epoch.create(epochname="OSN-T090/2022-09-23")

    assert (epoch.rawfits.count() == len(os.listdir(Path(iop4conf.datadir) / "raw" / "OSN-T090" / "2022-09-23")))
   
    epoch.build_master_biases()
    epoch.build_master_flats()

    assert (epoch.masterbias.count() > 0)
    assert (epoch.masterflats.count() > 0)


@pytest.mark.skip(reason="Not implemented yet")
@pytest.mark.django_db(transaction=True)
def test_polarimetry_groups(load_test_catalog):
    r""" Tests the splitting of polarimetry observations into groups.
    
    Organizing observations into groups is essential to derive polarimetry results.

    For OSN-T090 POLARIMETRY observations with AndorT090 instrument, four observations
    are needed to derive a single polarimetry result, for the same source, same band and same exptime,
    but different polarization angle.

    """

    assert False

@pytest.mark.django_db(transaction=True)
def test_build_single_proc(load_test_catalog):
    """ Test the whole building process of reduced fits in a single process """

    from iop4lib.db import Epoch, ReducedFit

    epochname_L = ["OSN-T090/2022-09-23", "OSN-T090/2022-09-18"]
    epoch_L = [Epoch.create(epochname=epochname, check_remote_list=False) for epochname in epochname_L]

    for epoch in epoch_L:
        epoch.build_master_biases()
        epoch.build_master_flats()

    iop4conf.nthreads = 1

    epoch = Epoch.by_epochname("OSN-T090/2022-09-18")
    
    epoch.reduce()

    assert (ReducedFit.objects.filter(epoch=epoch).count() == 1)

    for redf in ReducedFit.objects.filter(epoch=epoch).all():
        assert (redf.has_flag(ReducedFit.FLAGS.BUILT_REDUCED))
        assert not (redf.has_flag(ReducedFit.FLAGS.ERROR_ASTROMETRY))


@pytest.mark.django_db(transaction=True)
def test_build_multi_proc_photopol(load_test_catalog):
    """ Test the whole building process of reduced fits through multiprocessing 
    
    Also tests here relative photometry and polarimetry results and their 
    quality (value + uncertainties) (to avoid losing time reducing them
    in another test function).

    See build_test_dataset.py for the test dataset used in this test.

    """

    from iop4lib.db import Epoch, RawFit, ReducedFit
    from iop4lib.enums import IMGTYPES, SRCTYPES, INSTRUMENTS

    # get epochs that have Andor90 observations in them
    from iop4lib.iop4 import list_local_epochnames
    epochname_L = list_local_epochnames()
    epoch_L = [Epoch.create(epochname=epochname) for epochname in epochname_L]
    epoch_L = [epoch for epoch in epoch_L if epoch.rawfits.filter(instrument=INSTRUMENTS.AndorT90).exists()]

    for epoch in epoch_L:
        epoch.build_master_biases()
        epoch.build_master_flats()

    # 1. Test multi-process reduced fits building

    iop4conf.nthreads = 6

    rawfits = RawFit.objects.filter(epoch__in=epoch_L, instrument=INSTRUMENTS.AndorT90, imgtype=IMGTYPES.LIGHT).all()
    
    Epoch.reduce_rawfits(rawfits)

    assert (ReducedFit.objects.filter(epoch__in=epoch_L).count() == 5)

    for redf in ReducedFit.objects.filter(epoch__in=epoch_L).all():
        assert (redf.has_flag(ReducedFit.FLAGS.BUILT_REDUCED))
        assert not (redf.has_flag(ReducedFit.FLAGS.ERROR_ASTROMETRY))

    from iop4lib.db import PhotoPolResult, AstroSource

    # 2. Test relative photometry 

    logger.info("Testing relative photometry results")

    epoch = Epoch.by_epochname("OSN-T090/2022-09-18")

    epoch.compute_relative_photometry()

    qs_res = PhotoPolResult.objects.filter(epoch=epoch).all()

    # we expect only one photometry result target in this test dataset for this epoch
    assert qs_res.exclude(astrosource__in=AstroSource.objects.filter(is_calibrator=True)).count() == 1

    res = qs_res.get(astrosource__name="2200+420")

    # check that the photometric result is correct to 1.5 sigma or 0.02 mag compared to IOP3
    assert res.mag == approx(13.35, abs=max(1.5*res.mag_err, 0.05))

    # check that uncertainty of the photometric result is less than 0.06 mag
    assert res.mag_err < 0.06

    # 3. Test relative polarimetry

    logger.info("Testing relative polarimetry results")

    epoch = epoch.by_epochname("OSN-T090/2022-09-08")

    epoch.compute_relative_polarimetry()

    qs_res = PhotoPolResult.objects.filter(epoch=epoch).all()

    # we expect only one polarimetry result target in this test dataset for this epoch
    assert qs_res.exclude(astrosource__in=AstroSource.objects.filter(is_calibrator=True)).count() == 1

    res = qs_res.get(astrosource__name="2200+420")

    # check that the magnitude and polarization of the result relative to IOP3
    # for polarimetry, we expect a higher uncertainty than for photometry

    assert res.mag == approx(13.38, abs=max(1.5*res.mag_err, 0.05)) # 1.5 sigma or 0.05 mag
    assert res.mag_err < 0.1

    assert res.p == approx(14.0/100, abs=max(2*res.p_err, 2/100)) # 2 sigma or 2% of polarization degree
    assert res.p_err < 0.5/100

    assert res.chi == approx(14.7, abs=max(2*res.chi_err, 5)) # 2 sigma or 5 degrees of polarization angle
    assert res.chi_err < 2.5



# %% [markdown]
# # Notebook configuration
# Before using IOP4 inside a jupyter notebook, you need to configure the DB with the following lines

# %%
# %autoawait off
import iop4lib.config
iop4conf = iop4lib.Config(config_db=True, gonogui=False, jupytermode=True) 

# %% [markdown]
# Now that IOP4 is configured, you can import and use the DB models. E.g.:
# ```ipython
# from iop4lib.db import RawFit, ReducedFit
# ```

# %% [markdown]
# If you want to use the autoreload extension, e.g. with
# ```ipython
# %load_ext autoreload
# %autoreload all
# ```
# at the top of your notebook. You might encounter some problems when the modifications affect the DB models. In that case, re-executing the import statement a few times should help, otherwise you will need to restart the kernel.

# Since you are not using the `iop4.py` script, you will need to configure logging to suit your needs. E.g, to log to standard output, you can use:

# %%
import sys

# configure logging
import coloredlogs
import logging

logger = logging.getLogger()

# remove handlers

for handler in logger.handlers:
    logger.removeHandler(handler)

# set debug level and a nicer format for notebooks

logger.setLevel(logging.DEBUG)

logger_handler = logging.StreamHandler(stream=sys.stdout)
logger.addHandler(logger_handler)

iop4conf.log_date_format = '%Y-%m-%d %H:%M'
iop4conf.log_format = '%(asctime)s - pid %(process)d - [%(filename)s:%(lineno)d] - %(levelname)s - %(message)s'

logger_handler.setFormatter(coloredlogs.ColoredFormatter(iop4conf.log_format, datefmt=iop4conf.log_date_format))

# %% [markdown]
# Now you are ready to start using IOP4 inside your notebook.


from django.contrib import admin

from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.urls import path, reverse 
from django.http import HttpResponse, HttpResponseRedirect, HttpResponseNotFound
from django.shortcuts import render

# other imports
from iop4api.filters import *
from iop4api.models import *
from iop4api import views

# other imports
from iop4lib.enums import *
from astropy.time import Time

class AdminAperPhotResult(admin.ModelAdmin):
    model = AperPhotResult
    list_display = ['id', 'get_telescope', 'get_instrument', 'get_datetime', 'get_src_name', 'get_src_type', 'get_fwhm', 'get_aperpix', 'get_r_in', 'get_r_out', 'get_reducedfit', 'get_obsmode', 'pairs', 'get_rotangle', 'get_src_type', 'get_flux_counts', 'get_flux_counts_err', 'get_bkg_flux_counts', 'get_bkg_flux_counts_err', 'get_image_preview', 'modified']
    readonly_fields = [field.name for field in AperPhotResult._meta.fields]
    search_fields = ['id', 'reducedfit__instrument', 'astrosource__name', 'astrosource__srctype', 'reducedfit__id']
    list_filter = ['reducedfit__instrument', 'astrosource__srctype', 'reducedfit__epoch__telescope', 'reducedfit__obsmode']


    def get_urls(self):
        urls = super().get_urls()
        my_urls = [
            path('preview/<int:pk>', self.admin_site.admin_view(self.view_preview),  name=f"iop4api_{self.model._meta.model_name}_preview"),
        ]
        return my_urls + urls
    
    def view_preview(self, request, pk):
        
        if ((fit := self.model.objects.filter(id=pk).first()) is None):
            return HttpResponseNotFound()

        imgbytes = fit.get_img()
        return HttpResponse(imgbytes, content_type="image/png")
    
    @admin.display(description="TELESCOPE")
    def get_telescope(self, obj):
        return obj.reducedfit.epoch.telescope

    @admin.display(description="INSTRUMENT")
    def get_instrument(self, obj):
        return obj.reducedfit.instrument

    @admin.display(description="DATETIME")
    def get_datetime(self, obj):
        return Time(obj.reducedfit.juliandate, format='jd').strftime('%Y-%m-%d %H:%M:%S')
    
    @admin.display(description="SRCNAME")
    def get_src_name(self, obj):
        return obj.astrosource.name
    
    @admin.display(description="SRCTYPE")
    def get_src_type(self, obj):
        return obj.astrosource.srctype
    
    @admin.display(description="ReducedFit")
    def get_reducedfit(self, obj):
        self.allow_tags = True
        url = reverse('iop4admin:%s_%s_changelist' % (ReducedFit._meta.app_label, ReducedFit._meta.model_name)) + f"?id={obj.reducedfit.id}"
        return mark_safe(rf'<a href="{url}">{obj.reducedfit.id}</a>')    
    
    @admin.display(description="OBSMODE")
    def get_obsmode(self, obj):
        return obj.reducedfit.obsmode
    
    @admin.display(description="ROTANGLE")
    def get_rotangle(self, obj):
        return obj.reducedfit.rotangle
    
    @admin.display(description="fwhm")
    def get_fwhm(self, obj):
        if obj.fwhm is None:
            return "-"
        return f"{obj.fwhm:.1f}"
    
    @admin.display(description="aperpix")
    def get_aperpix(self, obj):
        if obj.aperpix is None:
            return "-"
        return f"{obj.aperpix:.1f}"
    
    @admin.display(description="r_in")
    def get_r_in(self, obj):
        if obj.r_in is None:
            return "-"
        return f"{obj.r_in:.1f}"
    
    @admin.display(description="r_out")
    def get_r_out(self, obj):
        if obj.r_out is None:
            return "-"
        return f"{obj.r_out:.1f}"
    
    @admin.display(description="flux_counts")
    def get_flux_counts(self, obj):
        if obj.flux_counts is None:
            return "-"
        else:
            return f"{obj.flux_counts:.1f}"
    
    @admin.display(description="flux_counts_err")
    def get_flux_counts_err(self, obj):
        if obj.flux_counts_err is None:
            return "-"
        else:
            return f"{obj.flux_counts_err:.1f}"
    
    @admin.display(description="bkg_flux_counts")
    def get_bkg_flux_counts(self, obj):
        if obj.bkg_flux_counts is None:
            return "-"
        else:
            return f"{obj.bkg_flux_counts:.1f}"
    
    @admin.display(description="bkg_flux_counts_err")
    def get_bkg_flux_counts_err(self, obj):
        if obj.bkg_flux_counts_err is None:
            return "-"
        else:
            return f"{obj.bkg_flux_counts_err:.1f}"
    
    @admin.display(description="img")
    def get_image_preview(self, obj, allow_tags=True):
        url_img_preview = reverse(f"iop4admin:iop4api_{self.model._meta.model_name}_preview", args=[obj.id])
        return format_html(rf"<img src='{url_img_preview}' width='64' height='64' />")


# Taken from
# https://github.com/cta-observatory/project-template-python-pure/blob/main/src/template/version.py

try:
    try:
        from ._dev_version import version
    except Exception:
        from ._version import version
except Exception:
    import warnings

    warnings.warn(
        "Could not determine version; this indicates a broken installation."
        " Install from PyPI, using conda or from a local git repository."
        " Installing github's autogenerated source release tarballs "
        " does not include version information and should be avoided."
    )
    del warnings
    version = "0.0.0"

__version__ = version



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""This script will create a iop4testdata.$MD5.tar.gz file in your home directory.

Use it to build a test dataset from your reduced and tested results. The .tar.gz 
will contain all the raw data needed to reduce the given photo-polarimetry results,
including raw light, bias, darks, flats and a test catalog file.

Edit the pk_L list to add the pk of the results you want to include in the test dataset.

Edit the raw_fileloc_L list to add the paths of additional files you want to include in the test dataset.

"""

import iop4lib.config
iop4conf = iop4lib.Config(config_db=True)

from iop4lib.db import Epoch, RawFit, ReducedFit, PhotoPolResult, MasterBias, MasterFlat, AperPhotResult, AstroSource
from iop4lib.telescopes import Telescope

import os
import subprocess
import shutil
import hashlib
from pathlib import Path


######################
### Configuration ####
######################


clean_workdir = False

## Addtional files to include 

Nmax_raw_per_master = 1

# only Nmax_raw_per_master will be used for these files

raw_fileloc_L = [
    # some dipol files to test calibration
    "OSN-T090/2023-11-06/BLLac_IAR-0001R.fit", # DIPIL photometry astrocalibration / shotgun
    "OSN-T090/2023-10-25/HD204827_R_IAR-0384.fts", # DIPOL polarimetry astrocalibration / target E, O in a star
    # DIPOL polarimetry astrocalibration / quad matching in a blazar
    "OSN-T090/2023-11-06/BLLac_IAR-0001R.fit", # the photometry file
    "OSN-T090/2023-11-06/BLLAC_R_IAR-0760.fts", # the polarimetry file
    "OSN-T090/2023-10-11/OJ248_R_IAR-0111.fts", # the polarimetry file
    "OSN-T090/2023-11-13/OJ248_R_full_IAR-0001R.fit", # the photometry file
]

# Results to include 

# For these, Nmax_raw_per_master will be ignored (to keep the result as close to the real one as possible)

# Since the pks of the results might have changed since last built, build the list of pks from the filelocs

results_filelocs = [
    "OSN-T090/2022-09-08/BLLac-0001R0.fit",

    # <PhotoPolResult(id: 287755                                                                                                                                                            
    #     reducedfits: [6565, 6566, 6567, 6568]
    #     POLARIMETRY R / 2200+420                                                                 
    #     JD: 2459831.46059 (2022-09-08 23:03:14.825)                                            
    #     mag: 13.346 ± 0.086  
    #     p: 0.125 ± 0.004                                                                       
    #     chi: 15.504 ± 0.909)> 

    "OSN-T090/2022-09-18/BLLac-0001R.fit", 

    # <PhotoPolResult(id: 284969
    #     reducedfits: [39071]                                                                   
    #     PHOTOMETRY R / 2200+420
    #     JD: 2459841.46187 (2022-09-18 23:05:05.190)                                            
    #     mag: 13.310 ± 0.032)>

    "CAHA-T220/2022-09-18/caf-20220918-23:01:33-sci-agui.fits",

    # <PhotoPolResult(id: 285934
    #     reducedfits: [40827, 40828, 40829, 40830]                                                                                                                                         
    #     POLARIMETRY R / 2200+420 
    #     JD: 2459841.46034 (2022-09-18 23:02:53.250)
    #     mag: 13.338 ± 0.026
    #     p: 0.110 ± 0.002     
    #     chi: 25.180 ± 0.409)>   

    "OSN-T090/2023-10-09/3C345_R_IAR-0001.fts",

    # <PhotoPolResult(id: 239179
    #     reducedfits: [105740, 105741, 105742, 105743, 105744, 105745, 105746, 105747, 105748, 105749, 105750, 105751, 105752, 105753, 105754, 105755]
    #     DIPOL POLARIMETRY R / 1641+399
    #     JD: 2460227.33940 (2023-10-09 20:08:43.813)
    #     p: 0.301 ± 0.020
    #     chi: 45.556 ± 1.841)>

    "CAHA-T220/2023-10-09/caf-20231009-19:17:38-sci-agui.fits",

    # <PhotoPolResult(id: 238940
    #     reducedfits: [57754, 57755, 57756, 57757]
    #     POLARIMETRY R / 1641+399
    #     JD: 2460227.30780 (2023-10-09 19:23:13.500)
    #     mag: 16.537 ± 0.013
    #     p: 0.293 ± 0.003
    #     chi: 46.617 ± 0.457)>

    "OSN-T090/2023-10-25/HD204827_R_IAR-0369.fts",

    # <PhotoPolResult(id: 236438
    #     reducedfits: [100574, 100575, 100576, 100577, 100578, 100579, 100580, 100581, 100582, 100583, 100584, 100585, 100586, 100587, 100588, 100589]
    #     DIPOL POLARIMETRY R / HD 204827
    #     JD: 2460243.55174 (2023-10-26 01:14:30.563)
    #     p: 0.051 ± 0.005
    #     chi: 61.374 ± 2.836)>
]

pk_L = list(set([PhotoPolResult.objects.filter(astrosource__in=AstroSource.objects.filter(is_calibrator=False), reducedfits__in=[ReducedFit.by_fileloc(fileloc)]).first().id for fileloc in results_filelocs]))


output_file = Path("~/iop4testdata.tar.gz").expanduser()
workdir = Path("~/iop4testdata").expanduser()

# remove workdir if configured, and create it again if necessary

if clean_workdir and os.path.exists(workdir):
    shutil.rmtree(workdir)

os.makedirs(workdir, exist_ok=True)


###############################################
### Files for the PhotoPolResults pk_L list ###
###############################################


print("Copying files for the PhotoPolResults pk_L list")

# get the list of files needed and copy them to the workdir

rawfitL = list()

for pk in pk_L:
    res = PhotoPolResult.objects.get(id=pk)

    if res.instrument == "DIPOL":
        Nmax = 1 # too many files otherwise
    else:
        Nmax = None

    for redf in res.reducedfits.all():
        # append all rawfits needed for this reducedfit
        rawfitL.append(redf.rawfit)

        # the bias for its masterbias
        for bias in redf.masterbias.rawfits.all()[:Nmax]:
            rawfitL.append(bias)

        # the darks for its masterdark
        if redf.masterdark is not None:
            for dark in redf.masterdark.rawfits.all()[:Nmax]:
                rawfitL.append(dark)

        # the flats for its masterflat
        for flat in redf.masterflat.rawfits.all()[:Nmax]:
            rawfitL.append(flat)

        # the bias for the masterbias of its masterflat
        for bias in redf.masterflat.masterbias.rawfits.all()[:Nmax]:
            rawfitL.append(bias)

        # the bias for the masterbias of its masterdark
        if redf.masterdark is not None:
            for bias in redf.masterdark.masterbias.rawfits.all()[:Nmax]:
                rawfitL.append(bias)

        # the bias and the darks for the masterdark of its masterflat
        if redf.masterflat.masterdark is not None:
            for dark in redf.masterflat.masterdark.rawfits.all()[:Nmax]:
                rawfitL.append(dark)


files_to_download = set([rawfit for rawfit in rawfitL if not os.path.exists(rawfit.filepath)])

if len(files_to_download) > 0:
    # download the files if needed
    Telescope.get_by_name(rawfitL[0].epoch.telescope).download_rawfits(files_to_download)

rawfitL = list(set(rawfitL))

print(f"Copying {len(rawfitL)} files for the given PhotoPolResults")

for rawfit in rawfitL:

    print(f"  cp {rawfit.fileloc}")

    relative_path = Path(rawfit.filepath).relative_to(iop4conf.datadir)
    
    dest = workdir / relative_path

    if not os.path.exists(os.path.dirname(dest)):
        os.makedirs(os.path.dirname(dest))

    if not os.path.exists(dest):
        shutil.copy(rawfit.filepath, dest)



###############################
### create the catalog file ###
###############################

print("Creating the catalog file")

astrosources_ids_L = list()

for res in PhotoPolResult.objects.filter(id__in=pk_L):
    astrosources_ids_L.append(res.astrosource.id)

    for calibrator in AstroSource.objects.filter(calibrates=res.astrosource):
        astrosources_ids_L.append(calibrator.id)

for fileloc in raw_fileloc_L:
    rf = RawFit.by_fileloc(fileloc)
    if (src := rf.header_hintobject) is not None:
        astrosources_ids_L.append(src.id)
        for calibrator in AstroSource.objects.filter(calibrates=src):
            astrosources_ids_L.append(calibrator.id)

for astrosource in AstroSource.objects.filter(id__in=set(astrosources_ids_L)):
    print(f"  {astrosource}")

from django.core.serializers import serialize
with open(workdir / "testcatalog.yaml", "a") as f:
    f.write(serialize("yaml", AstroSource.objects.filter(id__in=set(astrosources_ids_L)), use_natural_foreign_keys=True, use_natural_primary_keys=True))
    f.write('\n')



#############################
### copy additional files ###
#############################

print(f"Copying additional {len(raw_fileloc_L)} files")

rawfitL = list()

for fileloc in raw_fileloc_L:

    redf = ReducedFit.by_fileloc(fileloc)

    # append all rawfits needed for this reducedfit
    rawfitL.append(redf.rawfit)

    # the bias for its masterbias
    for bias in redf.masterbias.rawfits.all()[:Nmax_raw_per_master]:
        rawfitL.append(bias)

    # the darks for its masterdark
    if redf.masterdark is not None:
        for dark in redf.masterdark.rawfits.all()[:Nmax_raw_per_master]:
            rawfitL.append(dark)

    # the flats for its masterflat
    for flat in redf.masterflat.rawfits.all()[:Nmax_raw_per_master]:
        rawfitL.append(flat)

    # the bias for the masterbias of its masterflat
    for bias in redf.masterflat.masterbias.rawfits.all()[:Nmax_raw_per_master]:
        rawfitL.append(bias)

    # the bias for the masterbias of its masterdark
    if redf.masterdark is not None:
        for bias in redf.masterdark.masterbias.rawfits.all()[:Nmax_raw_per_master]:
            rawfitL.append(bias)

    # the bias and the darks for the masterdark of its masterflat
    if redf.masterflat.masterdark is not None:
        for dark in redf.masterflat.masterdark.rawfits.all()[:Nmax_raw_per_master]:
            rawfitL.append(dark)


for rawfit in rawfitL:

    print(f"  cp {rawfit.fileloc}")

    relative_path = Path(rawfit.filepath).relative_to(iop4conf.datadir)
    
    dest = workdir / relative_path

    if not os.path.exists(os.path.dirname(dest)):
        os.makedirs(os.path.dirname(dest))

    if not os.path.exists(dest):
        shutil.copy(rawfit.filepath, dest)
        

###########################
### Create .tar.gz file ###
###########################

print("Creating .tar.gz file")

# in mac-os, additional ._* files are created to save xattrs, avoid it setting COPYFILE_DISABLE
os.environ["COPYFILE_DISABLE"] = "true"

result = subprocess.run(["tar", "--no-xattrs", "--exclude='.*'", "-czf", str(output_file), "-C", str(workdir.parent), str(workdir.name)])

if result.returncode != 0:
    print("Error creating the .tar.gz file")
    exit(1)

print(" Created {}".format(output_file))

# read it and give the md5 sum

with open(output_file, "rb") as f:
    md5 = hashlib.md5(f.read()).hexdigest()
    print("{} {}".format(output_file.name, md5))

# add the md5sum to the filename
os.rename(output_file, output_file.parent / f"iop4testdata.{md5}.tar.gz")

print(f"{output_file.parent / f'iop4testdata.{md5}.tar.gz'}")

exit(0)


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports
from django.db import models
from django.forms.models import model_to_dict

# other imports
import os
import math
import numpy as np
import astropy.units as u
from astropy.wcs import WCS
import astropy.io.fits as fits

# iop4lib imports
from iop4lib.telescopes import Telescope
from iop4lib.instruments import Instrument
from iop4lib.utils.filedproperty import FiledProperty
from iop4lib.enums import *
from .rawfit import RawFit
from .astrosource import AstroSource
from .photopolresult import PhotoPolResult
from .aperphotresult import AperPhotResult

# logging
import logging
logger = logging.getLogger(__name__)

    
class ReducedFit(RawFit):
    """
    Reduced fits (bias and flat corrected, with wcs coordinates, etc, source list, etc).
    """

    # Database fields and information.

    # identifier
    rawfit = models.OneToOneField('RawFit', on_delete=models.CASCADE, related_name='reduced', parent_link=True, help_text="RawFit of this ReducedFit.")

    # ReducedFit specific fields
    masterbias = models.ForeignKey('MasterBias', null=True, on_delete=models.CASCADE, related_name='reduced', help_text="MasterBias to be used for the reduction.")
    masterflat = models.ForeignKey('MasterFlat', null=True, on_delete=models.CASCADE, related_name='reduced', help_text="MasterFlat to be used for the reduction.")
    masterdark = models.ForeignKey('MasterDark', null=True, on_delete=models.CASCADE, related_name='reduced', help_text="MasterDark to be used for the reduction.")
    sources_in_field = models.ManyToManyField('AstroSource', related_name='in_reducedfits', blank=True, help_text="Sources in the field of this FITS.")
    modified = models.DateTimeField(auto_now=True, help_text="Last time this entry was modified.")

    class Meta(RawFit.Meta):
        app_label = "iop4api"
        verbose_name = "ReducedFit"
        verbose_name_plural = "Reduced FITS files"

    # Properties

    @property
    def filepath(self):
       """ Returns the path to the file.
       The reduced FITS files are stored in the .filedpropdir directory (the same .filedpropdir as the corresponding RawFit).
       """
       return os.path.join(self.filedpropdir, self.filename)
    
    # Filed properties
    astrometry_info = FiledProperty()

    # Constructors

    @classmethod
    def create(cls, rawfit, 
               masterbias=None,
               masterflat=None,
               masterdark=None,
               auto_build=False,
               force_rebuild=False,
               auto_merge_to_db=True):
        """Creates a new instance of ReducedFit from a RawFit, updates existing entry if it exists.

        Parameters
        ----------
        rawfit : RawFit
            The RawFit to be reduced.
        masterbias : MasterBias, optional
            The MasterBias to be used for the reduction. If None, an appropiate MasterBias will be found.
        masterflat : MasterFlat, optional
            The MasterFlat to be used for the reduction. If None, an appropiate MasterFlat will be found.

        Other Parameters
        ----------------
        auto_build : bool, optional
            If auto_build is True, the file will be built if it does not exist.
        force_rebuild : bool, optional
            If force_rebuild is True, the file will be built regardless of whether it exists or not; this takes 
            precedence over auto_build.

        Returns
        -------
        None

        Notes
        -----
            See ReducedFit.build_file for more information about how the file is build.
        """
        from iop4lib.db import RawFit

        if rawfit is None:
            raise ValueError("rawfit must be provided.")
        
        if rawfit.imgtype != IMGTYPES.LIGHT:
            raise ValueError("Only RawFits of type LIGHT can be reduced.")
        
        if (reduced := ReducedFit.objects.filter(rawfit=rawfit).first()) is None:
            logger.debug(f"Creating DB entry for reduced fit {rawfit.fileloc}")     
            # get fields from rawfit, else they get overwritten
            # TODO: check if there is a better way to do this
            #field_dict = model_to_dict(rawfit)
            field_dict = RawFit.objects.filter(id=rawfit.id).values().get()
            logger.debug(f"Trying to create ReducedFit from RawFit, {field_dict=}. Check code if you encounter any problem. Might give an error if some foreign field is stored in rawfit. TODO: might be better to use .save_base()?")
            reduced = cls(rawfit=rawfit, **field_dict)
            reduced.save()
        else:
            logger.debug(f"DB entry of ReducedFit for {rawfit.fileloc} already exists, it will be used instead.")

        # instance only attributes
        reduced.auto_merge_to_db = auto_merge_to_db

        reduced.associate_masters(masterbias=masterbias, masterdark=masterdark, masterflat=masterflat)

        # build file

        if (not reduced.fileexists and auto_build) or force_rebuild:
            logger.info(f"Building file")
            reduced.build_file()

        # merge to db

        if auto_merge_to_db:
            reduced.save()

        return reduced
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.auto_merge_to_db = True
    

    # REDUCTION METHODS

    ## Delegated to telescopes or instrument classes

    def associate_masters(self, *args, **kwargs):
        return Instrument.by_name(self.instrument).associate_masters(self, *args, **kwargs)

    def apply_masters(self):
        return Instrument.by_name(self.instrument).apply_masters(self)
    
    def build_file(self, **build_wcs_kwargs):
        return Instrument.by_name(self.instrument).build_file(self, **build_wcs_kwargs)

    def astrometric_calibration(self, **build_wcs_kwargs):
        return Instrument.by_name(self.instrument).astrometric_calibration(self, **build_wcs_kwargs)

    @property
    def has_pairs(self):
        """ Indicates whether both ordinary and extraordinary sources are present in the file. """
        return Instrument.by_name(self.instrument).has_pairs(self)
        
    @property
    def wcs(self):
        """ Returns the WCS of the reduced fit. """
        return WCS(self.header, key="A")
    
    @property
    def wcs1(self):
        """ Returns the WCS of the first extension of the reduced fit. """
        return WCS(self.header, key="A")
    
    @property
    def wcs2(self):
        """ Returns the WCS of the second extension of the reduced fit. """
        return WCS(self.header, key="B")

    @property
    def pixscales(self):
        """ Pixel scales of the reduced fit as quantities (deg/pix) (it returns both the x and y scales)."""
        pix_scales_deg_pix = np.sqrt(np.sum(self.wcs.pixel_scale_matrix**2, axis=0))
        return pix_scales_deg_pix * u.Unit("deg / pix")
    
    @property
    def pixscale(self):
        """ Pixel scale of the reduced fit, in units of deg/pix (it returns the mean of x and y scales so it is strictly correct only for square pixels). """
        return  np.mean(self.pixscales)
    
    @property
    def centercoord(self):
        return self.wcs.pixel_to_world(self.mdata.shape[0]//2, self.mdata.shape[1]//2)

    @property
    def pixscale_equiv(self):
        """ Returns an equivalencies between pixels and angles for this reduced fit. """
        return u.pixel_scale(self.pixscale)
    
    @property
    def header_hintcoord(self):
        return self.rawfit.header_hintcoord
    
    @property
    def header_hintobject(self):
        return self.rawfit.header_hintobject

    def get_astrometry_position_hint(self, allsky=False, n_field_width=1.5, hintsep=None):
        return Instrument.by_name(self.instrument).get_astrometry_position_hint(self.rawfit, allsky=allsky,  n_field_width=n_field_width, hintsep=hintsep)
    
    def get_astrometry_size_hint(self):
        return Instrument.by_name(self.instrument).get_astrometry_size_hint(self.rawfit)
    
    def compute_aperture_photometry(self, *args, **kwargs):
        """ Delegated to the instrument. """
        return Instrument.by_name(self.instrument).compute_aperture_photometry(self, *args, **kwargs)

    def compute_relative_photometry(self, *args, **kwargs):
        """ Delegated to the instrument. """
        return Instrument.by_name(self.instrument).compute_relative_photometry(self, *args, **kwargs)
    
    @classmethod
    def compute_relative_polarimetry(cls, polarimetry_group, *args, **kwargs):
        """ Delegated to the instrument. """
        
        if not all([redf.telescope == polarimetry_group[0].telescope for redf in polarimetry_group]):
            raise Exception("All reduced fits in a polarimetry group must be from the same telescope")
        
        if not all([redf.instrument == polarimetry_group[0].instrument for redf in polarimetry_group]):
            raise Exception("All reduced fits in a polarimetry group must be from the same instrument")
        
        return Instrument.by_name(polarimetry_group[0].telescope).compute_relative_polarimetry(polarimetry_group, *args, **kwargs)